"""
Rotations contains all rotation like MoveGenerator classes.

.. inheritance-diagram:: fullrmc.Generators.Rotations
    :parts: 1


+-----------------------------------------------------+-----------------------------------------------------+
|.. figure:: randomRotation.png                       |.. figure:: randomRotationAboutAxis.png              |
|   :width: 375px                                     |   :width: 375px                                     |
|   :height: 300px                                    |   :height: 300px                                    |
|   :align: left                                      |   :align: left                                      |
|                                                     |                                                     |
|   Random rotation axis and angle generated and      |   Random rotation generated about a pre-defined axis|
|   applied on a Tetrahydrofuran molecule. Solid      |   or one of the symmetry axes of the Tetrahydrofuran|
|   colours are of the origin molecule position while |   molecule. Solid colours are of the origin molecule|
|   fading ones are of the rotated molecule.          |   position while fading ones are of the rotated     |
|   (:class:`RotationGenerator`)                      |   molecule.                                         |
|                                                     |   (:class:`RotationAboutAxisGenerator`              |
|                                                     |   :class:`RotationAboutSymmetryAxisGenerator`)      |
+-----------------------------------------------------+-----------------------------------------------------+
|.. figure:: orientationGenerator.png                 |                                                     |
|   :width: 375px                                     |                                                     |
|   :height: 300px                                    |                                                     |
|   :align: left                                      |                                                     |
|                                                     |                                                     |
|   Random orientation of hexane molecule to [1,1,1]  |                                                     |
|   axis with maximumOffsetAngle of 10 degrees is     |                                                     |
|   generated. First principal axis of hexane molecule|                                                     |
|   is considered as groupAxis. Solid colors are of   |                                                     |
|   original molecule while fading ones are of the    |                                                     |
|   oriented one. (:class:`OrientationGenerator`)     |                                                     |
|                                                     |                                                     |
+-----------------------------------------------------+-----------------------------------------------------+

.. raw:: html

    <iframe width="560" height="315"
    src="https://www.youtube.com/embed/-clLvYiaC8w?rel=0"
    frameborder="0" allowfullscreen>
    </iframe>

"""
# standard libraries imports
from __future__ import print_function

# external libraries imports
import numpy as np

# fullrmc imports
from ..Globals import INT_TYPE, FLOAT_TYPE, PI, PRECISION, LOGGER
from ..Globals import str, long, unicode, bytes, basestring, range, xrange, maxint
from ..Core.Collection import is_number, is_integer, get_path, generate_random_float, get_principal_axis, get_rotation_matrix, orient, generate_vectors_in_solid_angle, generate_random_vector
from ..Core.MoveGenerator import MoveGenerator, PathGenerator


class RotationGenerator(MoveGenerator):
    """
    Generate random rotational moves upon groups of atoms.
    Only groups of more than one atom are accepted.

    :Parameters:
        #. group (None, Group): The group instance.
        #. amplitude (number): The maximum rotation angle allowed in degrees.
           It must be strictly bigger than 0 and strictly smaller than 360.

    .. code-block:: python

        # import fullrmc modules
        from fullrmc.Engine import Engine
        from fullrmc.Generators.Rotations import RotationGenerator

        # create engine
        ENGINE = Engine(path='my_engine.rmc')

        # set pdb file
        ENGINE.set_pdb('system.pdb')

        # Add constraints ...
        # Re-define groups if needed ...
        # Re-define groups selector if needed ...

        # set moves generators to random rotations.
        # Maximum rotation amplitude is set to 5 degrees to all defined groups
        for g in ENGINE.groups:
            if len(g) >1:
                g.set_move_generator( RotationGenerator(amplitude=5) )

    """
    def __init__(self, group=None, amplitude=2):
        super(RotationGenerator, self).__init__(group=group)
        # set amplitude
        self.set_amplitude(amplitude)

    @property
    def amplitude(self):
        """ Maximum allowed angle of rotation in rad."""
        return self.__amplitude

    def set_amplitude(self, amplitude):
        """
        Set maximum rotation angle in degrees and transforms it to rad.

        :Parameters:
            #. amplitude (number): the maximum allowed rotation angle in
               degrees. It must be strictly bigger than 0 and strictly smaller
               than 360.
        """
        assert is_number(amplitude), LOGGER.error("rotation amplitude must be a number")
        amplitude = float(amplitude)
        assert amplitude>0, LOGGER.error("rotation amplitude must be bigger than 0 deg.")
        assert amplitude<360, LOGGER.error("rotation amplitude must be smaller than 360 deg.")
        # convert to radian and store amplitude
        self.__amplitude = FLOAT_TYPE(PI*amplitude/180.)

    def check_group(self, group):
        """
        Check the generator's group.

        :Parameters:
            #. group (Group): the Group instance.
        """
        if len(group.indexes)<=1:
            return False, "At least two atoms needed in a group to perform rotation."
        else:
            return True, ""

    def transform_coordinates(self, coordinates, argument=None):
        """
        Rotate coordinates.

        :Parameters:
            #. coordinates (np.ndarray): The coordinates on which to apply
               the rotation.
            #. argument (object): Any python object. Not used in this
               generator.

        :Returns:
            #. coordinates (np.ndarray): The new coordinates after applying
               the rotation.
        """
        if coordinates.shape[0]<=1:
            # atoms where removed, fall back to random translation
            return coordinates+generate_random_vector(minAmp=self.__amplitude[0],
                                                      maxAmp=self.__amplitude[1])
        else:
            # get rotation axis
            n = 0
            while n<PRECISION:
                rotationAxis = 1-2*np.random.random(3)
                n = np.linalg.norm(rotationAxis)
            rotationAxis /= n
            # get rotation angle
            rotationAngle = (1-2*generate_random_float())*self.amplitude
            # get rotation matrix
            rotationMatrix = get_rotation_matrix(rotationAxis, rotationAngle)
            # get atoms group center
            center = np.sum(coordinates, 0)/coordinates.shape[0]
            # translate to origin
            rotatedCoordinates = coordinates-center
            # rotate
            for idx in range(rotatedCoordinates.shape[0]):
                rotatedCoordinates[idx,:] = np.dot( rotationMatrix, rotatedCoordinates[idx,:])
            # translate back to center and return rotated coordinates
            return np.array(rotatedCoordinates+center, dtype=FLOAT_TYPE)


class RotationAboutAxisGenerator(RotationGenerator):
    """
    Generates random rotational moves upon groups of atoms about a
    pre-defined axis.

    :Parameters:
        #. group (None, Group): The group instance.
        #. amplitude (number): The maximum allowed rotation angle in degrees.
           It must be strictly bigger than 0 and strictly smaller than 360.
        #. axis (list,set,tuple,numpy.ndarray): The rotational axis vector.

    .. code-block:: python

        # import fullrmc modules
        from fullrmc.Engine import Engine
        from fullrmc.Generators.Rotations import RotationAboutAxisGenerator

        # create engine
        ENGINE = Engine(path='my_engine.rmc')

        # set pdb file
        ENGINE.set_pdb('system.pdb')

        # Add constraints ...
        # Re-define groups if needed ...
        # Re-define groups selector if needed ...

        # set moves generators to random rotations about (1,1,1) a pre-defined axis.
        # Maximum rotation amplitude is set to 5 degrees to all defined groups
        for g in ENGINE.groups:
            if len(g) >1:
                g.set_move_generator( RotationAboutAxisGenerator(amplitude=5, axis=(1,1,1)) )

    """

    def __init__(self, group=None, amplitude=2, axis=(1,0,0)):
        super(RotationAboutAxisGenerator, self).__init__(group=group, amplitude=amplitude)
        # set amplitude
        self.set_axis(axis)

    @property
    def axis(self):
        """ Rotation axis vector."""
        return self.__axis

    def check_group(self, group):
        """
        Check the generator's group.

        :Parameters:
            #. group (Group): The Group instance.
        """
        return True, ""

    def set_axis(self, axis):
        """
        Set the axis along which the rotation will be performed.

        :Parameters:
            #. axis (list,set,tuple,numpy.ndarray): The rotation axis vector.
        """
        assert isinstance(axis, (list,set,tuple,np.ndarray)), LOGGER.error("axis must be a list")
        axis = list(axis)
        assert len(axis)==3, LOGGER.error("axis list must have 3 items")
        for pos in axis:
            assert is_number(pos), LOGGER.error("axis items must be numbers")
        axis = [FLOAT_TYPE(pos) for pos in axis]
        axis =  np.array(axis, dtype=FLOAT_TYPE)
        self.__axis = axis/FLOAT_TYPE( np.linalg.norm(axis) )

    def transform_coordinates(self, coordinates, argument=None):
        """
        Rotate coordinates.

        :Parameters:
            #. coordinates (np.ndarray): The coordinates on which to apply
               the rotation.
            #. argument (object): Not used here.

        :Returns:
            #. coordinates (np.ndarray): The new coordinates after
               applying the rotation.
        """
        # get rotation angle
        rotationAngle  = (1-2*generate_random_float())*self.amplitude
        # get rotation matrix
        rotationMatrix = get_rotation_matrix(self.__axis, rotationAngle)
        # get atoms group center and rotation axis
        center,_,_,_,_,_,_ = get_principal_axis(coordinates)
        # translate to origin
        rotatedCoordinates = coordinates-center
        # rotate
        for idx in range(rotatedCoordinates.shape[0]):
            rotatedCoordinates[idx,:] = np.dot( rotationMatrix, rotatedCoordinates[idx,:])
        # translate back to center and return rotated coordinates
        return np.array(rotatedCoordinates+center, dtype=FLOAT_TYPE)


class RotationAboutSymmetryAxisGenerator(RotationGenerator):
    """
    Generates random rotational moves upon groups of atoms about one of
    their symmetry axis. Only groups of more than 1 atom are accepted.

    :Parameters:
        #. group (None, fullrmc.Engine): The constraint fullrmc engine.
        #. amplitude (number): Maximum rotation angle in degrees.
           It must be strictly bigger than 0 and strictly smaller than 360.
        #. axis (integer): Must be 0,1 or 2 for respectively the main,
           secondary or tertiary symmetry axis

    .. code-block:: python

        # import fullrmc modules
        from fullrmc.Engine import Engine
        from fullrmc.Generators.Rotations import RotationAboutSymmetryAxisGenerator

        # create engine
        ENGINE = Engine(path='my_engine.rmc')

        # set pdb file
        ENGINE.set_pdb('system.pdb')

        # Add constraints ...
        # Re-define groups if needed ...
        # Re-define groups selector if needed ...

        # set moves generators to random rotations about the second symmetry axis of each group.
        # Maximum rotation amplitude is set to 5 degrees to all defined groups
        for g in ENGINE.groups:
            if len(g) >1:
                g.set_move_generator( RotationAboutSymmetryAxisGenerator(amplitude=5, axis=1) )

    """
    def __init__(self, group=None, amplitude=2, axis=0):
        super(RotationAboutSymmetryAxisGenerator, self).__init__(group=group, amplitude=amplitude)
        # set amplitude
        self.set_axis(axis)

    @property
    def axis(self):
        """ Rotation axis index."""
        return self.__axis

    def set_axis(self, axis):
        """
        Set the symmetry axis index to rotate about.

        :Parameters:
            #. axis (integer): Must be 0,1 or 2 for respectively the main,
               secondary or tertiary symmetry axis.
        """
        assert is_integer(axis), LOGGER.error("rotation symmetry axis must be an integer")
        axis = INT_TYPE(axis)
        assert axis>=0, LOGGER.error("rotation symmetry axis must be positive.")
        assert axis<=2, LOGGER.error("rotation symmetry axis must be smaller or equal to 2")
        # convert to radian and store amplitude
        self.__axis = axis

    def transform_coordinates(self, coordinates, argument=None):
        """
        Rotate coordinates.

        :Parameters:
            #. coordinates (np.ndarray): The coordinates on which to apply
               the rotation.
            #. argument (object): Not used here.

        :Returns:
            #. coordinates (np.ndarray): The new coordinates after
               applying the rotation.
        """
        if coordinates.shape[0]<=1:
            # atoms where removed, fall back to random translation
            return coordinates+generate_random_vector(minAmp=self.__amplitude[0],
                                                      maxAmp=self.__amplitude[1])
        else:
            # get rotation angle
            rotationAngle = (1-2*generate_random_float())*self.amplitude
            # get atoms group center and rotation axis
            center,_,_,_,X,Y,Z =get_principal_axis(coordinates)
            rotationAxis = [X,Y,Z][self.__axis]
            # get rotation matrix
            rotationMatrix = get_rotation_matrix(rotationAxis, rotationAngle)
            # translate to origin
            rotatedCoordinates = coordinates-center
            # rotate
            for idx in range(rotatedCoordinates.shape[0]):
                rotatedCoordinates[idx,:] = np.dot( rotationMatrix, rotatedCoordinates[idx,:])
            # translate back to center and return rotated coordinates
            return np.array(rotatedCoordinates+center, dtype=FLOAT_TYPE)


class RotationAboutSymmetryAxisPath(PathGenerator):
    """
    Generate rotational moves upon groups of atoms about one of their
    symmetry axis. Only groups of more than one atom are accepted.

    :Parameters:
        #. group (None, fullrmc.Engine): The constraint fullrmc engine.
        #. axis (integer): Must be 0,1 or 2 for respectively the main,
           secondary or tertiary symmetry axis.
        #. path (List): list of angles.
        #. randomize (boolean): Whether to pull moves randomly from path or
           pull moves in order at every step.

    .. code-block:: python

        # import fullrmc modules
        from fullrmc.Engine import Engine
        from fullrmc.Generators.Rotations import RotationAboutSymmetryAxisPath

        # create engine
        ENGINE = Engine(path='my_engine.rmc')

        # set pdb file
        ENGINE.set_pdb('system.pdb')

        # Add constraints ...
        # Re-define groups if needed ...
        # Re-define groups selector if needed ...

        # set moves generators to pre-defined rotations about the second symmetry axis of each group.
        angles = [-0.1, -0.5, -0.05, 0.5, 0.01, 2, 3, 1, -3]
        for g in ENGINE.groups:
            if len(g) >1:
                g.set_move_generator( RotationAboutSymmetryAxisPath(axis=1, path=angles) )

    """
    def __init__(self, group=None,  axis=0, path=None, randomize=False):
        # initialize PathGenerator
        PathGenerator.__init__(self, group=group, path=path, randomize=randomize)
        # set axis
        self.set_axis(axis)

    @property
    def axis(self):
        """ Rotation axis index."""
        return self.__axis

    def set_axis(self, axis):
        """
        Set the symmetry axis index to rotate about.

        :Parameters:
            #. axis (integer): Must be 0,1 or 2 for respectively the main,
               secondary or tertiary symmetry axis
        """
        assert is_integer(axis), LOGGER.error("rotation symmetry axis must be an integer")
        axis = INT_TYPE(axis)
        assert axis>=0, LOGGER.error("rotation symmetry axis must be positive.")
        assert axis<=2,LOGGER.error("rotation symmetry axis must be smaller or equal to 2")
        # convert to radian and store amplitude
        self.__axis = axis

    def check_path(self, path):
        """
        Check the generator's path.

        :Parameters:
            #. path (None, list): The list of moves.
        """
        if not isinstance(path, (list, set, tuple, np.ndarray)):
            return False, "path must be a list"
        path = list(path)
        if not len(path):
            return False, "path can't be empty"
        for angle in path:
            if not is_number(angle):
                return False, "path items must be numbers"
        return True, ""

    def normalize_path(self, path):
        """
        Transform all path angles to radian.

        :Parameters:
            #. path (list): The list of moves in degrees.

        :Returns:
            #. path (list): The list of moves in rad.
        """
        path = [FLOAT_TYPE(angle)*PI/FLOAT_TYPE(180.) for angle in path]
        return list(path)

    def check_group(self, group):
        """
        Check the generator's group.

        :Parameters:
            #. group (Group): The Group instance.
        """
        if len(group.indexes)<=1:
            return False, "At least two atoms needed in a group to perform rotation."
        else:
            return True, ""

    def transform_coordinates(self, coordinates, argument):
        """
        Rotate coordinates.

        :Parameters:
            #. coordinates (np.ndarray): The coordinates on which to apply
               the rotation.
            #. argument (object): The rotation angle.

        :Returns:
            #. coordinates (np.ndarray): The new coordinates after applying
               the rotation.
        """
        if coordinates.shape[0]<=1:
            # atoms where removed, fall back to random translation
            return coordinates+generate_random_vector(minAmp=self.__amplitude[0],
                                                      maxAmp=self.__amplitude[1])
        else:
            # get atoms group center and rotation axis
            center,_,_,_,X,Y,Z =get_principal_axis(coordinates)
            rotationAxis = [X,Y,Z][self.__axis]
            # get rotation matrix
            rotationMatrix = get_rotation_matrix(rotationAxis, argument)
            # translate to origin
            rotatedCoordinates = coordinates-center
            # rotate
            for idx in range(rotatedCoordinates.shape[0]):
                rotatedCoordinates[idx,:] = np.dot( rotationMatrix, rotatedCoordinates[idx,:])
            # translate back to center and return rotated coordinates
            return np.array(rotatedCoordinates+center, dtype=FLOAT_TYPE)



class OrientationGenerator(MoveGenerator):
    """
    Generate rotational moves upon groups of atoms to align and orient
    along an axis. Orientation rotations are computed randomly allowing
    offset angle between grouAxis and orientationAxis Only groups of more
    than 1 atom are accepted.

    :Parameters:
        #. group (None, Group): The group instance.
        #. maximumOffsetAngle (number): The maximum offset angle in degrees
           between groupAxis and orientationAxis.
        #. groupAxis (dict): The group axis. Only one key is allowed.
           If key is 'fixed', value must be a list, tuple or a numpy.array of
           a vector such as [X,Y,Z]. If key is 'symmetry', in this case the
           group axis is computed as one of the three symmetry axis of the
           group atoms. the value must be even 0, 1 or 2 for respectively
           the first, second and tertiary symmetry axis.
        #. orientationAxis (dict): The axis to align the group with.
           If key is 'fixed', value must be a list, tuple or a numpy.array
           of a vector such as [X,Y,Z]. If Key is 'symmetry', in this case
           the value must be a list of two items, the first one is a list
           of atoms indexes to compute symmetry axis and the second item
           must be even 0, 1 or 2 for respectively the first, second and
           tertiary symmetry axis.
        #. flip (None, bool): Whether to allow flipping axis orientation
           or not. If True, orientationAxis will be flipped forcing
           anti-parallel orientation. If False, orientationAxis will not be
           flipped forcing parallel orientation. If None is given, no
           flipping is forced, flipping can be set randomly to True or False
           during run time execution.

    .. code-block:: python

        # import fullrmc modules
        from fullrmc.Engine import Engine
        from fullrmc.Generators.Rotations import OrientationGenerator

        # create engine
        ENGINE = Engine(path='my_engine.rmc')

        # set pdb file
        ENGINE.set_pdb('system.pdb')

        # Add constraints ...
        # Re-define groups if needed ...
        # Re-define groups selector if needed ...

        # set moves generators to orientations of each group third symmetry axis
        # towards the (-1,0,2) the predefined axis within maximum 5 degrees.
        for g in ENGINE.groups:
            if len(g) >1:
                g.set_move_generator( OrientationGenerator(maximumOffsetAngle=5,
                                                           groupAxis={"symmetry":2},
                                                           orientationAxis={"fixed":(-1,0,2)}) )
    """
    def __init__(self, group=None, maximumOffsetAngle=10, groupAxis={"symmetry":0}, orientationAxis={"fixed":(1,0,0)}, flip=None):
        super(OrientationGenerator, self).__init__(group=group)
        # set maximumOffsetAngle
        self.set_maximum_offset_angle(maximumOffsetAngle)
        # set group axis
        self.set_group_axis(groupAxis)
        # set orientation axis
        self.set_orientation_axis(orientationAxis)
        # set flip
        self.set_flip(flip)

    @property
    def maximumOffsetAngle(self):
        """ Maximum offset angle in degrees between groupAxis and
        orientationAxis in rad."""
        return self.__maximumOffsetAngle

    @property
    def orientationAxis(self):
        """ Orientation axis value or definition."""
        return self.__orientationAxis

    @property
    def groupAxis(self):
        """ Group axis value or definition."""
        return self.__groupAxis

    @property
    def flip(self):
        """ Flip value."""
        return self.__flip

    def set_maximum_offset_angle(self, maximumOffsetAngle):
        """
        Set the maximum offset angle allowed.

        :Parameters:
            #. maximumOffsetAngle (number): The maximum offset angle in
               degrees between groupAxis and orientationAxis.
        """
        assert is_number(maximumOffsetAngle), LOGGER.error("maximumOffsetAngle must be a number")
        maximumOffsetAngle = float(maximumOffsetAngle)
        assert maximumOffsetAngle>0, LOGGER.error("maximumOffsetAngle must be bigger than 0 deg.")
        assert maximumOffsetAngle<180, LOGGER.error("maximumOffsetAngle must be smaller than 180 deg.")
        # convert to radian and store amplitude
        self.__maximumOffsetAngle = FLOAT_TYPE(PI*maximumOffsetAngle/180.)

    def check_group(self, group):
        """
        Check the generator's group.

        :Parameters:
            #. group (Group): the Group instance.
        """
        if len(group.indexes)<=1:
            return False, "At least two atoms needed in a group to perform rotation."
        else:
            return True, ""

    def set_flip(self, flip):
        """
        Set flip flag value.

        :Parameters:
            #. flip (None, bool): Whether to allow flipping axis orientation
               or not. If True, orientationAxis will be flipped forcing
               anti-parallel orientation. If False, orientationAxis will not
               be flipped forcing parallel orientation. If None is given,
               no flipping is forced, flipping can be set randomly to True
               or False during run time execution.
        """
        assert flip in (None, True, False), LOGGER.error("flip can only be None, True or False")
        self.__flip = flip

    def set_group_axis(self, groupAxis):
        """
        Sets group axis value.

        :Parameters:
           #. groupAxis (dict): The group axis. Only one key is allowed.
              If key is fixed, value must be a list, tuple or a numpy.array
              of a vector such as [X,Y,Z]. If key is symmetry, in this case
              the group axis is computed as one of the three symmetry axis of
              the group atoms. the value must be even 0, 1 or 2 for
              respectively the first, second and tertiary symmetry axis.
        """
        assert isinstance(groupAxis, dict), LOGGER.error("groupAxis must be a dictionary")
        assert len(groupAxis) == 1, LOGGER.error("groupAxis must have a single key")
        key = groupAxis[0]
        val = groupAxis[key]
        if key == "fixed":
            self.__mustComputeGroupAxis = False
            assert isinstance(val, (list,set,tuple,np.ndarray)), LOGGER.error("groupAxis value must be a list")
            if isinstance(val, np.ndarray):
                assert len(val.shape) == 1, LOGGER.error("groupAxis value must have a single dimension")
            val = list(val)
            assert len(val)==3, LOGGER.error("groupAxis fixed value must be a vector")
            for v in val:
                assert is_number(v), LOGGER.error("groupAxis value item must be numbers")
            val  = np.array([FLOAT_TYPE(v) for v in val], dtype=FLOAT_TYPE)
            norm = FLOAT_TYPE(np.sqrt(np.sum(val**2)))
            val /= norm
        elif key == "symmetry":
            self.__mustComputeGroupAxis = True
            assert is_integer(val), LOGGER.error("groupAxis symmetry value must be an integer")
            val = INT_TYPE(val)
            assert val>=0 and val<3, LOGGER.error("groupAxis symmetry value must be positive smaller than 3")
        else:
            self.__mustComputeGroupAxis = None
            raise Exception(LOGGER.error("groupAxis key must be either 'fixed' or 'symmetry'"))
        # set groupAxis
        self.__groupAxis = {key:val}

    def set_orientation_axis(self, orientationAxis):
        """
        Set orientation axis value.

        :Parameters:
            #. orientationAxis (dict): The axis to align the group axis with.
               If key is fixed, value must be a list, tuple or a numpy.array
               of a vector such as [X,Y,Z]. If Key is symmetry, in this case
               the value must be a list of two items, the first one is a list
               of atoms indexes to compute symmetry axis and the second item
               must be even 0, 1 or 2 for respectively the first, second and
               tertiary symmetry axis.
        """
        assert isinstance(orientationAxis, dict), LOGGER.error("orientationAxis must be a dictionary")
        assert len(orientationAxis) == 1, LOGGER.error("orientationAxis must have a single key")
        key = orientationAxis[0]
        val = orientationAxis[key]
        if key == "fixed":
            self.__mustComputeOrientationAxis = False
            assert isinstance(val, (list,set,tuple,np.ndarray)), LOGGER.error("orientationAxis value must be a list")
            if isinstance(val, np.ndarray):
                assert len(val.shape) == 1, LOGGER.error("orientationAxis value must have a single dimension")
            val = list(val)
            assert len(val)==3, LOGGER.error("orientationAxis fixed value must be a vector")
            for v in val:
                assert is_number(v), LOGGER.error("orientationAxis value item must be numbers")
            val  = np.array([FLOAT_TYPE(v) for v in val], dtype=FLOAT_TYPE)
            norm = FLOAT_TYPE(np.sqrt(np.sum(val**2)))
            val /= norm
        elif key == "symmetry":
            self.__mustComputeOrientationAxis = True
            assert isinstance(val, (list, tuple)), LOGGER.error("orientationAxis symmetry value must be a list")
            assert len(val) == 2, LOGGER.error("orientationAxis symmetry value must be a list of two items")
            val0 = []
            for v in val[0]:
                assert is_integer(v), LOGGER.error("orientationAxis symmetry value list items must be integers")
                v0 = INT_TYPE(v)
                assert v0>=0, LOGGER.error("orientationAxis symmetry value list items must be positive")
                val0.append(v0)
            assert len(set(val0))==len(val[0]), LOGGER.error("orientationAxis symmetry value list redundant items indexes found")
            val0 = np.array(val0, dtype=INT_TYPE)
            val1 = val[1]
            assert is_integer(val1), LOGGER.error("orientationAxis symmetry value second item must be an integer")
            val1 = INT_TYPE(val1)
            assert val1>=0 and val1<3, LOGGER.error("orientationAxis symmetry value second item must be positive smaller than 3")
            val = (val0,val1)
        else:
            self.__mustComputeOrientationAxis = None
            raise Exception(LOGGER.error("orientationAxis key must be either 'fixed' or 'symmetry'"))
        # set orientationAxis
        self.__orientationAxis = {key:val}

    def __get_orientation_axis__(self):
        if self.__mustComputeOrientationAxis:
            coordinates   = self.group._get_engine().realCoordinates[self.__orientationAxis["symmetry"][0]]
            _,_,_,_,X,Y,Z = get_principal_axis(coordinates)
            axis = [X,Y,Z][self.__orientationAxis["symmetry"][1]]
        else:
            axis  = self.__orientationAxis["fixed"]
        return axis

    def __get_group_axis__(self, coordinates):
        if self.__mustComputeGroupAxis:
            _,_,_,_,X,Y,Z = get_principal_axis(coordinates)
            axis = [X,Y,Z][self.__groupAxis["symmetry"]]
        else:
            axis = self.__groupAxis["fixed"]
        return axis

    def transform_coordinates(self, coordinates, argument=None):
        """
        Rotate coordinates.

        :Parameters:
            #. coordinates (np.ndarray): The coordinates on which to apply
               the rotation.
            #. argument (object): Not used here.

        :Returns:
            #. coordinates (np.ndarray): The new coordinates after applying
               the rotation.
        """
        if coordinates.shape[0]<=1:
            # atoms where removed, fall back to random translation
            return coordinates+generate_random_vector(minAmp=self.__amplitude[0],
                                                      maxAmp=self.__amplitude[1])
        else:
           # create flip flag
            if self.__flip is None:
                flip = FLOAT_TYPE( np.sign(1-2*generate_random_float()) )
            elif self.__flip:
                flip = FLOAT_TYPE(-1)
            else:
                flip = FLOAT_TYPE(1)
            # get group axis
            groupAxis = self.__get_group_axis__(coordinates)
            # get align axis within offset angle
            orientationAxis = flip*self.__get_orientation_axis__()
            orientationAxis = generate_vectors_in_solid_angle(direction=orientationAxis,
                                                              maxAngle=self.__maximumOffsetAngle,
                                                              numberOfVectors=1)[0]
            # get coordinates center
            center = np.array(np.sum(coordinates, 0)/coordinates.shape[0] , dtype=FLOAT_TYPE)
            # translate to origin
            rotatedCoordinates = coordinates-center
            # align coordinates
            rotatedCoordinates = orient(rotatedCoordinates, groupAxis, orientationAxis)
            # translate back to center and return rotated coordinates
            return np.array(rotatedCoordinates+center, dtype=FLOAT_TYPE)



##########################################################################################
##############################  IMPORTING USEFUL DEFINITIONS  ############################
# standard libraries imports
from __future__ import print_function
import os, sys, time

# external libraries imports
import numpy as np
import matplotlib.pyplot as plt

# fullrmc library imports
from fullrmc.Globals import LOGGER, maxint
from fullrmc.Engine import Engine
from fullrmc.Constraints.PairDistributionConstraints import PairDistributionConstraint
from fullrmc.Constraints.DistanceConstraints import InterMolecularDistanceConstraint
from fullrmc.Constraints.BondConstraints import BondConstraint
from fullrmc.Constraints.AngleConstraints import BondsAngleConstraint
from fullrmc.Constraints.ImproperAngleConstraints import ImproperAngleConstraint


##########################################################################################
##################################  SHUT DOWN LOGGING  ###################################
LOGGER.set_minimum_level(maxint, stdoutFlag=True, fileFlag=True)


##########################################################################################
#####################################  CREATE ENGINE  ####################################
# parameters
NSTEPS  = 10000
pdbPath = 'system.pdb'
expData = 'experimental.gr'
# initialize engine
ENGINE = Engine(path=None)
ENGINE.set_pdb(pdbPath)
# create constraints
PDF_CONSTRAINT = PairDistributionConstraint(experimentalData=expData, weighting="atomicNumber")
EMD_CONSTRAINT = InterMolecularDistanceConstraint()
B_CONSTRAINT   = BondConstraint()
BA_CONSTRAINT  = BondsAngleConstraint()
IA_CONSTRAINT  = ImproperAngleConstraint()
# add constraints to engine
ENGINE.add_constraints([PDF_CONSTRAINT, EMD_CONSTRAINT, B_CONSTRAINT, BA_CONSTRAINT, IA_CONSTRAINT])
# initialize constraints definitions
B_CONSTRAINT.create_bonds_by_definition( bondsDefinition={"THF": [('O' ,'C1' , 1.22, 1.70),
                                                                  ('O' ,'C4' , 1.22, 1.70),
                                                                  ('C1','C2' , 1.25, 1.90),
                                                                  ('C2','C3' , 1.25, 1.90),
                                                                  ('C3','C4' , 1.25, 1.90),
                                                                  ('C1','H11', 0.58, 1.22),('C1','H12', 0.58, 1.22),
                                                                  ('C2','H21', 0.58, 1.22),('C2','H22', 0.58, 1.22),
                                                                  ('C3','H31', 0.58, 1.22),('C3','H32', 0.58, 1.22),
                                                                  ('C4','H41', 0.58, 1.22),('C4','H42', 0.58, 1.22)] })
BA_CONSTRAINT.create_angles_by_definition( anglesDefinition={"THF": [ ('O'  ,'C1' ,'C4' , 105, 125),
                                                                      ('C1' ,'O'  ,'C2' , 100, 120),
                                                                      ('C4' ,'O'  ,'C3' , 100, 120),
                                                                      ('C2' ,'C1' ,'C3' , 95 , 115),
                                                                      ('C3' ,'C2' ,'C4' , 95 , 115),
                                                                      # H-C-H angle
                                                                      ('C1' ,'H11','H12', 98 , 118),
                                                                      ('C2' ,'H21','H22', 98 , 118),
                                                                      ('C3' ,'H31','H32', 98 , 118),
                                                                      ('C4' ,'H41','H42', 98 , 118),
                                                                      # H-C-O angle
                                                                      ('C1' ,'H11','O'  , 100, 120),
                                                                      ('C1' ,'H12','O'  , 100, 120),
                                                                      ('C4' ,'H41','O'  , 100, 120),
                                                                      ('C4' ,'H42','O'  , 100, 120),
                                                                      # H-C-C
                                                                      ('C1' ,'H11','C2' , 103, 123),
                                                                      ('C1' ,'H12','C2' , 103, 123),
                                                                      ('C2' ,'H21','C1' , 103, 123),
                                                                      ('C2' ,'H21','C3' , 103, 123),
                                                                      ('C2' ,'H22','C1' , 103, 123),
                                                                      ('C2' ,'H22','C3' , 103, 123),
                                                                      ('C3' ,'H31','C2' , 103, 123),
                                                                      ('C3' ,'H31','C4' , 103, 123),
                                                                      ('C3' ,'H32','C2' , 103, 123),
                                                                      ('C3' ,'H32','C4' , 103, 123),
                                                                      ('C4' ,'H41','C3' , 103, 123),
                                                                      ('C4' ,'H42','C3' , 103, 123) ] })
IA_CONSTRAINT.create_angles_by_definition( anglesDefinition={"THF": [ ('C2','O','C1','C4', -15, 15),
                                                                      ('C3','O','C1','C4', -15, 15) ] })


##########################################################################################
####################################  DIFFERENT RUNS  ####################################
def run(nsteps, groups=None, pdf=False, vdw=False, bond=False, angle=False, improper=False, message=""):
    # reset pdb
    ENGINE.set_pdb(pdbPath)
    ENGINE.set_groups(groups)
    # initialize constraints data
    PDF_CONSTRAINT.set_used(pdf)
    EMD_CONSTRAINT.set_used(vdw)
    B_CONSTRAINT.set_used(bond)
    BA_CONSTRAINT.set_used(angle)
    IA_CONSTRAINT.set_used(improper)
    # run
    ENGINE.initialize_used_constraints()
    tic = time.time()
    ENGINE.run(numberOfSteps=nsteps, saveFrequency=2*nsteps, restartPdb=None)
    spentTime = time.time()-tic
    # log
    message += "constraints --> ("
    if pdf:
        message += "pdf,"
    if vdw:
        message += "vdw,"
    if bond:
        message += "bond,"
    if angle:
        message += "angle,"
    if improper:
        message += "improper,"
    message += ") steps (%i) time (%i sec) \n"%(nsteps, spentTime)
    LOGGER.info("generated:%i - tried:%i - accepted:%i"%(ENGINE.generated, ENGINE.tried, ENGINE.accepted))
    LOGGER.info(message)
    return float(spentTime)/float(nsteps), ENGINE.tried , ENGINE.accepted

def benchmark_constraints(groupsList):
    print("================ Benchmark constraints ================")
    # run benchmark
    benchmark = {}
    accepted  = {}
    tried     = {}
    for GN in groupsList:
        print("++++++++ %i atoms per group"%GN)
        groups = [np.array(item, dtype=np.int32) for item in zip( *[range(idx,ENGINE.numberOfAtoms,GN) for idx in range(GN)] )]
        benchmark[GN] = {}
        tried[GN]     = {}
        accepted[GN]  = {}
        print("---- No constraints")
        benchmark[GN]['0001_none'],     tried[GN]['0001_none'] ,     accepted[GN]['0001_none']     = run(nsteps=NSTEPS, groups=groups, pdf=False, vdw=False, bond=False, angle=False, improper=False, message="atoms (%i) "%GN)
        print("---- PairDistributionConstraint")
        benchmark[GN]['0002_pdf'],      tried[GN]['0002_pdf'] ,      accepted[GN]['0002_pdf']      = run(nsteps=NSTEPS, groups=groups, pdf=True,  vdw=False, bond=False, angle=False, improper=False, message="atoms (%i) "%GN)
        print("---- InterMolecularDistanceConstraint")
        benchmark[GN]['0003_vdw'],      tried[GN]['0003_vdw'] ,      accepted[GN]['0003_vdw']      = run(nsteps=NSTEPS, groups=groups, pdf=False, vdw=True,  bond=False, angle=False, improper=False, message="atoms (%i) "%GN)
        print("---- BondConstraint")
        benchmark[GN]['0004_bond'],     tried[GN]['0004_bond'] ,     accepted[GN]['0004_bond']     = run(nsteps=NSTEPS, groups=groups, pdf=False, vdw=False, bond=True,  angle=False, improper=False, message="atoms (%i) "%GN)
        print("---- BondsAngleConstraint")
        benchmark[GN]['0005_angle'],    tried[GN]['0005_angle'] ,    accepted[GN]['0005_angle']    = run(nsteps=NSTEPS, groups=groups, pdf=False, vdw=False, bond=False, angle=True,  improper=False, message="atoms (%i) "%GN)
        print("---- ImproperAngleConstraint")
        benchmark[GN]['0006_improper'], tried[GN]['0006_improper'] , accepted[GN]['0006_improper'] = run(nsteps=NSTEPS, groups=groups, pdf=False, vdw=False, bond=False, angle=False, improper=True,  message="atoms (%i) "%GN)
        print("---- All constraints")
        benchmark[GN]['0007_all'],      tried[GN]['0007_all'] ,      accepted[GN]['0007_all']      = run(nsteps=NSTEPS, groups=groups, pdf=True,  vdw=True,  bond=True,  angle=True,  improper=True,  message="atoms (%i) "%GN)

    # plot
    plt.figure()
    atoms = sorted(benchmark)
    bench = [np.array(atoms)]
    accep = [np.array(atoms)]
    tri   = [np.array(atoms)]
    header = "groups "
    for key in sorted(benchmark[atoms[0]]):
        times = np.array( [ benchmark[n][key] for n in atoms] )
        header += " %s"%key.split("_")[1]
        bench.append(times)
        tri.append( np.array( [ tried[n][key] for n in atoms] ) )
        accep.append( np.array( [ accepted[n][key] for n in atoms] ) )
        plt.plot(atoms, times, label=key.split("_")[1])
    # annotate tried(accepted)
    for i, txt in enumerate( accep[-1] ):
        plt.gca().annotate(str(tri[-1][i])+"("+str(txt)+")", (bench[0][i],bench[-1][i]))
    # show plot
    plt.legend()
    plt.title("Constraints Benchmark")
    plt.xlabel("number of atoms")
    plt.ylabel("time per step (s)")
    # save
    np.savetxt(fname="benchmark_constraints_time.dat",     X=np.transpose(bench), fmt='%.10f', delimiter='    ', newline='\n', header=header)
    np.savetxt(fname="benchmark_constraints_tried.dat",    X=np.transpose(tri),   fmt='%.10f', delimiter='    ', newline='\n', header=header)
    np.savetxt(fname="benchmark_constraints_accepted.dat", X=np.transpose(accep), fmt='%.10f', delimiter='    ', newline='\n', header=header)


def benchmark_nsteps(constraint, groupSize=13, stepsList=range(5000,105000,5000)):
    print("================ Benchmark number of steps ================")
    CS = {'pdf':False, 'vdw':False, 'bond':False, 'angle':False, 'improper':False}
    if not constraint in CS:
        if constraint == "all":
            for k in CS:
                CS[k]=True
        else:
            assert constraint == "none"
    print( "++++++++ Used constraints are %s"%CS)
    # set groups
    groups = [np.array(item, dtype=np.int32) for item in zip( *[range(idx,ENGINE.numberOfAtoms,groupSize) for idx in range(groupSize)] )]
    # run benchmark
    benchmark = []
    accepted  = []
    tried     = []
    stepsList = sorted(stepsList)
    for ns in stepsList:
        print("---- %i steps"%ns)
        bench, tri, accep = run(nsteps=ns, groups=groups, pdf=CS['pdf'], vdw=CS['vdw'], bond=CS['bond'], angle=CS['angle'], improper=CS['improper'], message="atoms (%i) "%groupSize)
        benchmark.append( bench )
        tried.append( tri )
        accepted.append( accep )

    # create data
    benchmark = [np.array(stepsList), np.array(benchmark)]
    tried     = [np.array(stepsList), np.array(tried)]
    accepted  = [np.array(stepsList), np.array(accepted)]
    # plot
    plt.figure()
    plt.plot(benchmark[0], benchmark[1])
    # annotate tried(accepted)
    for i, txt in enumerate( accepted[-1] ):
        plt.gca().annotate(str(tried[-1][i])+"("+str(txt)+")", (benchmark[0][i],benchmark[-1][i]))
    # show plot
    plt.title("Number of steps Benchmark")
    plt.xlabel("number of steps")
    plt.ylabel("time per step (s)")
    # save
    np.savetxt(fname='benchmark_%sSteps_%iGroupSize_time.dat'%(constraint,groupSize),     X=np.transpose(benchmark), fmt='%.10f', delimiter='    ', newline='\n', header="steps timePerStep(s)")
    np.savetxt(fname='benchmark_%sSteps_%iGroupSize_tried.dat'%(constraint,groupSize),    X=np.transpose(tried),     fmt='%.10f', delimiter='    ', newline='\n', header="steps timePerStep(s)")
    np.savetxt(fname='benchmark_%sSteps_%iGroupSize_accepted.dat'%(constraint,groupSize), X=np.transpose(accepted),  fmt='%.10f', delimiter='    ', newline='\n', header="steps timePerStep(s)")

def load_and_plot_constraints_benchmark():
    benchmark = np.loadtxt(fname='benchmark_constraints_time.dat')
    tried     = np.loadtxt(fname='benchmark_constraints_tried.dat')
    accepted  = np.loadtxt(fname='benchmark_constraints_accepted.dat')
    # plot
    keys = open('benchmark_constraints_time.dat').readlines()[0].split("#")[1].split()
    minY = 0
    maxY = 0
    for idx in range(1,len(keys)):
        style = '-'
        if keys[idx] == 'all':
            style = '.-'
        # mean accepted
        meanAccepted = int( sum(benchmark[:,0]*accepted[:,idx])/sum(benchmark[:,0]) )
        plt.plot(benchmark[:,0], benchmark[:,idx], style, label=keys[idx]+" (%i)"%meanAccepted)
        minY = min(minY, min(benchmark[:,idx]) )
        maxY = max(minY, max(benchmark[:,idx]) )
    # annotate tried(accepted)
    for i, txt in enumerate( accepted[:,-1] ):
        T = 100*float(tried[i,-1])/float(tried[1,1])
        A = 100*float(accepted[i,-1])/float(tried[1,1])
        plt.gca().annotate( "%.2f%% (%.2f%%)"%(T,A),  #"%i (%i)"%( int(tried[i,-1]),int(txt) ),
                            xy = (benchmark[i,0],benchmark[i,-1]),
                            rotation=90,
                            horizontalalignment='center',
                            verticalalignment='bottom')
    # show plot
    plt.legend(frameon=False, loc='upper left')
    plt.xlabel("Number of atoms per group")
    plt.ylabel("Time per step (s)")
    plt.gcf().patch.set_facecolor('white')
    # set fig size
    #figSize = plt.gcf().get_size_inches()
    #figSize[1] = figSize[1]+figSize[1]/2.
    #plt.gcf().set_size_inches(figSize, forward=True)
    plt.ylim((None, maxY+0.3*(maxY-minY)))
    # save
    plt.savefig("benchmark_constraint.png")
    # plot
    plt.show()

def load_and_plot_steps_benchmark(constraint="all", groupSize=13):
    benchmark = np.loadtxt(fname='benchmark_%sSteps_%iGroupSize_time.dat'%(constraint,groupSize) )
    tried     = np.loadtxt(fname='benchmark_%sSteps_%iGroupSize_tried.dat'%(constraint,groupSize) )
    accepted  = np.loadtxt(fname='benchmark_%sSteps_%iGroupSize_accepted.dat'%(constraint,groupSize) )
    # plot benchmark
    plt.plot(benchmark[:,0], benchmark[:,1])
    minY = min(benchmark[:,1])
    maxY = max(benchmark[:,1])
    # annotate tried(accepted)
    for i, txt in enumerate( accepted[:,-1] ):
        T = 100*float(tried[i,-1])/float(benchmark[i,0])
        A = 100*float(accepted[i,-1])/float(benchmark[i,0])
        plt.gca().annotate( "%.2f%% (%.2f%%)"%(T,A),  #str(int(tried[i,-1]))+" ("+str(int(txt))+")",
                            xy = (benchmark[i,0],benchmark[i,-1]),
                            rotation=90,
                            horizontalalignment='center',
                            verticalalignment='bottom')
    # show plot
    plt.legend(frameon=False, loc='upper left')
    plt.xlabel("Number of steps")
    plt.ylabel("Time per step (s)")
    plt.gcf().patch.set_facecolor('white')
    # set fig size
    #figSize = plt.gcf().get_size_inches()
    #figSize[1] = figSize[1]+figSize[1]/2.
    #plt.gcf().set_size_inches(figSize, forward=True)
    plt.ylim((None, maxY+0.3*(maxY-minY)))
    # save
    plt.savefig("benchmark_steps.png")
    # plot
    plt.show()



##########################################################################################
#####################################  RUN BENCHMARKS  ###################################
benchmark_constraints( range(1,30,1) )
benchmark_nsteps(constraint = 'all', groupSize=13, stepsList=range(5000,105000,5000))
# show benchmark plots
plt.show()

##########################################################################################
####################################  PLOT BENCHMARKS  ###################################
#load_and_plot_steps_benchmark()
#load_and_plot_constraints_benchmark()



"""
This is a C compiled module to compute bonded atoms angle.
"""      
from libc.math cimport sqrt, fabs
import cython
cimport cython
import numpy as np
cimport numpy as np
from numpy cimport ndarray
from fullrmc.Core.pairs_distances import  pair_difference_to_point

# declare types
NUMPY_FLOAT32 = np.float32
NUMPY_INT32   = np.int32
ctypedef np.float32_t C_FLOAT32
ctypedef np.int32_t   C_INT32

# declare constants
cdef C_FLOAT32 FLOAT_NEG_ONE   = -1.0
cdef C_FLOAT32 FLOAT_ZERO      = 0.0
cdef C_FLOAT32 FLOAT_ONE       = 1.0
cdef C_FLOAT32 FLOAT_TWO       = 2.0
cdef C_FLOAT32 BOX_LENGTH      = 1.0
cdef C_FLOAT32 HALF_BOX_LENGTH = 0.5
cdef C_INT32   INT_ZERO        = 0
cdef C_INT32   INT_ONE         = 1


cdef extern from "math.h":
    C_FLOAT32 floor(C_FLOAT32 x)
    C_FLOAT32 ceil(C_FLOAT32 x)
    C_FLOAT32 sqrt(C_FLOAT32 x)

cdef inline C_FLOAT32 round(C_FLOAT32 num):
    return floor(num + HALF_BOX_LENGTH) if (num > FLOAT_ZERO) else ceil(num - HALF_BOX_LENGTH)
    
       
       
@cython.nonecheck(False)
@cython.boundscheck(False)
@cython.wraparound(False)
@cython.cdivision(True)
@cython.always_allow_keywords(False)
cdef _single_angle( ndarray[C_FLOAT32, ndim=1] leftVectors,
                    ndarray[C_FLOAT32, ndim=1] rightVectors,
                    ndarray[C_FLOAT32, ndim=1] lowerLimits ,
                    ndarray[C_FLOAT32, ndim=1] upperLimits ,
                    ndarray[C_FLOAT32, ndim=1] angles ,
                    ndarray[C_FLOAT32, ndim=1] reducedAngles ,
                    C_INT32                    index,
                    bint                       reduceAngleToUpper = False,
                    bint                       reduceAngleToLower = False):
    """
    Computes the angles constraint given bonded atoms vectors.
    
    :Arguments:
       #. leftVectors (float32 array): The left vectors array.
       #. rightVectors (float32 array): The right vectors array.
       #. lowerLimits (float32 array): The (numberOfLeftIndexes) array for lower limit or minimum bond length allowed.
       #. upperLimits (float32 array): The (numberOfLeftIndexes) array for upper limit or maximum bond length allowed.
       #. angles (float32 array): The calculated angles (rad).
       #. reducedAngles (float32 array): The reduced angles (rad).
       #. index (int): index in lowerLimits, upperLimits, angles, reducedAngles arrays.
       #. reduceAngleToUpper (bool): Whether to reduce angle found out of limits to the difference between the angle and the upper limit. When True, this flag has the higher priority. DEFAULT: False
       #. reduceAngleToLower (bool): Whether to reduce angle found out of limits to the difference between the angle and the lower limit. When True, this flag may lose its priority for reduceAngleToUpper if the later is True. DEFAULT: False
    """
    # declare variables
    cdef C_FLOAT32 leftNorm, rightNorm, dot
    cdef C_FLOAT32 angle, reducedAngle
    cdef C_FLOAT32 lower, upper
    cdef C_FLOAT32 leftVector_x, leftVector_y, leftVector_z
    cdef C_FLOAT32 rightVector_x, rightVector_y, rightVector_z

    # compute left vector norm
    leftVector_x = leftVectors[0]
    leftVector_y = leftVectors[1]
    leftVector_z = leftVectors[2]
    leftNorm     = sqrt(leftVector_x*leftVector_x + leftVector_y*leftVector_y + leftVector_z*leftVector_z)
    if leftNorm==0:
        raise Exception("Computing angle, left vector found to have null length")
    # compute right vector norm
    rightVector_x = rightVectors[0]
    rightVector_y = rightVectors[1]
    rightVector_z = rightVectors[2]
    rightNorm     = sqrt(rightVector_x*rightVector_x + rightVector_y*rightVector_y + rightVector_z*rightVector_z)
    if rightNorm==0:
        raise Exception("Computing angle, right vector found to have null length")
    # compute dot product
    dot = leftVector_x*rightVector_x + leftVector_y*rightVector_y + leftVector_z*rightVector_z
    # calculate angle
    dot  /= (leftNorm*rightNorm)
    angle = np.arccos( np.clip( dot ,-1, 1 ) )  # np.arccos( dot ) clip for floating errors
    # compute reduced angle
    lower = lowerLimits[index]
    upper = upperLimits[index]
    if angle>=lower and angle<=upper:
        reducedAngle = FLOAT_ZERO     
    elif reduceAngleToUpper:
        reducedAngle = fabs(upper-angle)
    elif reduceAngleToLower:
        reducedAngle = fabs(lower-angle)
    else:
        if angle > (lower+upper)/FLOAT_TWO:
            reducedAngle = fabs(upper-angle)
        else:
            reducedAngle = fabs(lower-angle)
    # set angles and reduced
    angles[index]        = angle
    reducedAngles[index] = reducedAngle

    
    
    
@cython.nonecheck(False)
@cython.boundscheck(False)
@cython.wraparound(False)
@cython.cdivision(True)
@cython.always_allow_keywords(False)
def full_angles_coords( ndarray[C_INT32, ndim=1]   central not None,
                        ndarray[C_INT32, ndim=1]   left not None,
                        ndarray[C_INT32, ndim=1]   right not None,
                        ndarray[C_FLOAT32, ndim=1] lowerLimit not None,
                        ndarray[C_FLOAT32, ndim=1] upperLimit not None,
                        ndarray[C_FLOAT32, ndim=2] boxCoords not None,
                        ndarray[C_FLOAT32, ndim=2] basis not None,
                        bint                       isPBC,
                        bint                       reduceAngleToUpper = False,
                        bint                       reduceAngleToLower = False,
                        C_INT32                    ncores = 1):    
    """
    Computes the angles constraint given bonded atoms vectors.
    
    :Arguments:
       #. central (int32 (n,) numpy.ndarray): The central atom indexes.
       #. left (int32 (n,) numpy.ndarray): The left atom indexes.
       #. right (int32 (n,) numpy.ndarray): The right atom indexes.
       #. lowerLimit (float32 (n,) numpy.ndarray): The angles lower limits.
       #. upperLimit (float32 (n,) numpy.ndarray): The angles upper limits.
       #. boxCoords (float32 (n,3) numpy.ndarray): The atomic coordinates array of the same shape as pointsFrom.
       #. basis (float32 (3,3) numpy.ndarray): The (3x3) boundary conditions box vectors.
       #. isPBC (bool): Whether it is a periodic boundary conditions or infinite.
       #. reduceAngleToUpper (bool): Whether to reduce angle found out of limits to the difference between the angle and the upper limit. When True, this flag has the higher priority. DEFAULT: False
       #. reduceAngleToLower (bool): Whether to reduce angle found out of limits to the difference between the angle and the lower limit. When True, this flag may lose its priority for reduceAngleToUpper if the later is True. DEFAULT: False
       #. ncores (int32) [default=1]: The number of cores to use.
       
    :Returns:
       #. angles (float32 (n,) numpy.ndarray): The calculated angles (rad).
       #. reducedAngles (float32 (n,) numpy.ndarray): The reduced angles (rad).
    """
    cdef C_INT32 i, numberOfIndexes
    cdef C_FLOAT32 angle, reducedAngle
    # get number of indexes
    numberOfIndexes = <C_INT32>len(lowerLimit)
    # create abgles and reduced list
    cdef ndarray[C_FLOAT32,  mode="c", ndim=1] angles  = np.zeros((numberOfIndexes), dtype=NUMPY_FLOAT32)
    cdef ndarray[C_FLOAT32,  mode="c", ndim=1] reduced = np.zeros((numberOfIndexes), dtype=NUMPY_FLOAT32) 
    # loop all angles
    for i from 0 <= i < numberOfIndexes:
        leftVectors = pair_difference_to_point( point1 = boxCoords[central[i],:], 
                                                point2 = boxCoords[left[i],:], 
                                                basis  = basis,
                                                isPBC  = isPBC,
                                                ncores = INT_ONE)  
        rightVectors = pair_difference_to_point( point1 = boxCoords[central[i],:], 
                                                 point2 = boxCoords[right[i],:],
                                                 basis  = basis,
                                                 isPBC  = isPBC,
                                                 ncores = INT_ONE)                                                                                  
        _single_angle( leftVectors        = leftVectors , 
                       rightVectors       = rightVectors , 
                       lowerLimits        = lowerLimit ,
                       upperLimits        = upperLimit ,
                       angles             = angles,
                       reducedAngles      = reduced,
                       index              = i,
                       reduceAngleToUpper = reduceAngleToUpper,
                       reduceAngleToLower = reduceAngleToLower)
    # return results
    return angles, reduced















    


##########################################################################################
##############################  IMPORTING USEFUL DEFINITIONS  ############################
# standard libraries imports
import time, sys, os

# external libraries imports
import numpy as np
from pdbparser.pdbparser import pdbparser

# fullrmc library imports
from fullrmc.Globals import LOGGER
from fullrmc.Engine import Engine
from fullrmc.Constraints.DistanceConstraints import InterMolecularDistanceConstraint
from fullrmc.Generators.Translations import TranslationGenerator
from fullrmc.Generators.Rotations import RotationGenerator
from fullrmc.Core.Collection import generate_random_float
from fullrmc.Core.MoveGenerator import MoveGeneratorCollector
from fullrmc.Core.GroupSelector import RecursiveGroupSelector
from fullrmc.Selectors.RandomSelectors import RandomSelector
from fullrmc.Core.boundary_conditions_collection import transform_coordinates


##########################################################################################
###############################  BIASED ENGINE DEFINTION  ################################
class BiasedEngine(Engine):
    """
    This is a biased engine that fakes the computation of biasedStdErr
    just for the purpose of this example
    """
    def __init__(self, improveProbability=0.05, initialStdErr=1000000, *args, **kwargs):
        super(BiasedEngine, self).__init__(*args, **kwargs)
        self.__improveProbability = improveProbability
        self.biasedStdErr = initialStdErr
    
    def compute_total_standard_error(self, constraints, current=True):
        if generate_random_float() <= self.__improveProbability:
            newStdErr = self.biasedStdErr-generate_random_float()*1e-3
            if newStdErr<=0:
                raise Exception("biasedStdErr reached 0. restart using bigger initialStdErr")
        else:
            newStdErr = self.biasedStdErr+generate_random_float()*1e-3    
        return newStdErr
    
    def run(self, numberOfSteps=100000, saveFrequency=1000, savePath="restart", 
                  xyzFrequency=None, xyzPath="trajectory.xyz"):
        """
        This is an exact copy of engine run method with slight changes marked with #-->
        to make two trajectories, one of the real system and another the explored space.
        new code is marked with #<--
        all leading variables double scores __ removed.
        """
        # get arguments
        #-->_numberOfSteps            = self.__runtime_get_number_of_steps(numberOfSteps)
        #-->_saveFrequency, _savePath = self.__runtime_get_save_engine(saveFrequency, savePath)
        #-->_xyzFrequency, _xyzPath   = self.__runtime_get_save_xyz(xyzFrequency, xyzPath)
        _numberOfSteps = numberOfSteps #<--
        _saveFrequency = 2*numberOfSteps #<--
        _savePath = savePath  #<--
        # create xyz file
        #-->if _xyzFrequency is not None:
        #-->    _xyzfd = open(_xyzPath, 'a')
        _xyzfd = open("trajectory.xyz", 'a')#<--
        # get and initialize used constraints
        _usedConstraints, _constraints, _rigidConstraints = self.initialize_used_constraints()
        if not len(_usedConstraints):
            LOGGER.warn("No constraints are used. Configuration will be randomize")
        # compute biasedStdErr
        self.biasedStdErr = self.compute_total_standard_error(_constraints, current=True)
        # initialize useful arguments
        _engineStartTime    = time.time()
        _lastSavedChiSquare = self.biasedStdErr
        _coordsBeforeMove   = None
        _moveTried          = False
        # initialize group selector
        self.groupSelector._runtime_initialize()
        
        self.__realCoords = self.realCoordinates  #<--
        self.__boxCoords  = self.boxCoordinates  #<--
        #   #####################################################################################   #
        #   #################################### RUN ENGINE #####################################   #
        LOGGER.info("Engine started %i steps, biasedStdErr is: %.6f"%(_numberOfSteps, self.biasedStdErr) )
        self.__generated = 0 #<--
        self.__tried = 0     #<--
        self.__accepted=0    #<--
        for step in xrange(_numberOfSteps):
            # increment generated
            self.__generated += 1
            # get group
            self.__lastSelectedGroupIndex = self.groupSelector.select_index()
            group = self.groups[self.__lastSelectedGroupIndex]
            # get atoms indexes
            groupAtomsIndexes = group.indexes
            # get move generator
            groupMoveGenerator = group.moveGenerator
            # get group atoms coordinates before applying move 
            if _coordsBeforeMove is None or not self.groupSelector.isRecurring:
                _coordsBeforeMove = np.array(self.realCoordinates[groupAtomsIndexes], dtype=self.realCoordinates.dtype)
            elif self.groupSelector.explore:
                if _moveTried:
                    _coordsBeforeMove = movedRealCoordinates
            elif not self.groupSelector.refine:
                _coordsBeforeMove = np.array(self.realCoordinates[groupAtomsIndexes], dtype=self.realCoordinates.dtype)
            # compute moved coordinates
            movedRealCoordinates = groupMoveGenerator.move(_coordsBeforeMove)
            movedBoxCoordinates  = transform_coordinates(transMatrix=self.reciprocalBasisVectors , coords=movedRealCoordinates)
            ########################### compute enhanceOnlyConstraints ############################
            rejectMove = False
            for c in _rigidConstraints:
                # compute before move
                c.compute_before_move(realIndexes=groupAtomsIndexes, relativeIndexes=groupAtomsIndexes)
                # compute after move
                c.compute_after_move(realIndexes=groupAtomsIndexes, relativeIndexes=groupAtomsIndexes, movedBoxCoordinates=movedBoxCoordinates)
                # get rejectMove
                rejectMove = c.should_step_get_rejected(c.afterMoveStandardError)
                if rejectMove:
                    break
            _moveTried = not rejectMove
            ############################## reject move before trying ##############################
            if rejectMove:
                # enhanceOnlyConstraints reject move
                for c in _rigidConstraints:
                    c.reject_move(realIndexes=groupAtomsIndexes, relativeIndexes=groupAtomsIndexes)
                # log generated move rejected before getting tried
                LOGGER.log("move not tried","Generated move %i is not tried"%self.tried)
            ###################################### try move #######################################
            else:
                self.__tried += 1
                for c in _constraints:
                    # compute before move
                    c.compute_before_move(realIndexes=groupAtomsIndexes, relativeIndexes=groupAtomsIndexes)
                    # compute after move
                    c.compute_after_move(realIndexes=groupAtomsIndexes, relativeIndexes=groupAtomsIndexes, movedBoxCoordinates=movedBoxCoordinates)
            ################################ compute new biasedStdErr ################################
                newStdErr = self.compute_total_standard_error(_constraints, current=False)
                #if len(_constraints) and (newStdErr >= self.biasedStdErr):
                if newStdErr > self.biasedStdErr:
                    if generate_random_float() > self.tolerance:
                        rejectMove = True
                    else:
                        self.tolerated += 1
                        self.biasedStdErr  = newStdErr
                else:
                    self.biasedStdErr = newStdErr
            ################################## reject tried move ##################################
            if rejectMove:
                # set selector move rejected
                self.groupSelector.move_rejected(self.__lastSelectedGroupIndex)
                if _moveTried:
                    # constraints reject move
                    for c in _constraints:
                        c.reject_move(realIndexes=groupAtomsIndexes, relativeIndexes=groupAtomsIndexes)
                    # log tried move rejected
                    LOGGER.log("move rejected","Tried move %i is rejected"%self.__generated)
            ##################################### accept move #####################################
            else:
                self.__accepted  += 1
                # set selector move accepted
                self.groupSelector.move_accepted(self.__lastSelectedGroupIndex)
                # constraints reject move
                for c in _usedConstraints:
                    c.accept_move(realIndexes=groupAtomsIndexes, relativeIndexes=groupAtomsIndexes)
                # set new coordinates
                self.__realCoords[groupAtomsIndexes] = movedRealCoordinates
                self.__boxCoords[groupAtomsIndexes]  = movedBoxCoordinates
                # log new successful move
                triedRatio    = 100.*(float(self.__tried)/float(self.__generated))
                acceptedRatio = 100.*(float(self.__accepted)/float(self.__generated))
                LOGGER.log("move accepted","Generated:%i - Tried:%i(%.3f%%) - Accepted:%i(%.3f%%) - biasedStdErr:%.6f" %(self.__generated , self.__tried, triedRatio, self.__accepted, acceptedRatio, self.biasedStdErr))
            ##################################### save engine #####################################
            if _saveFrequency is not None:
                if not(step+1)%_saveFrequency:
                    if _lastSavedChiSquare==self.biasedStdErr:
                        LOGGER.info("Save engine omitted because no improvement made since last save.")
                    else:
                        # update state
                        self.state  = time.time()
                        for c in _usedConstraints:
                           #c.increment_tried()
                           c.set_state(self.state)
                        # save engine
                        _lastSavedChiSquare = self.biasedStdErr
                        self.save(_savePath)
            ############################### dump coords to xyz file ###############################
            #-->if _xyzFrequency is not None:
            #-->    if not(step+1)%_xyzFrequency:
            #-->        _xyzfd.write("%s\n"%self.__pdb.numberOfAtoms)
            #-->        triedRatio    = 100.*(float(self.__tried)/float(self.__generated))
            #-->        acceptedRatio = 100.*(float(self.__accepted)/float(self.__generated))
            #-->        _xyzfd.write("Generated:%i - Tried:%i(%.3f%%) - Accepted:%i(%.3f%%) - biasedStdErr:%.6f\n" %(self.__generated , self.__tried, triedRatio, self.__accepted, acceptedRatio, self.biasedStdErr))
            #-->        frame = [self.allNames[idx]+ " " + "%10.5f"%self.__realCoords[idx][0] + " %10.5f"%self.__realCoords[idx][1] + " %10.5f"%self.__realCoords[idx][2] + "\n" for idx in self.__pdb.xindexes]
            #-->        _xyzfd.write("".join(frame)) 
            triedRatio    = 100.*(float(self.__tried)/float(self.__generated)) #<--
            acceptedRatio = 100.*(float(self.__accepted)/float(self.__generated)) #<--
            _xyzfd.write("%s\n"%(len(groupAtomsIndexes)*2) ) #<--
            _xyzfd.write("Generated:%i - Tried:%i(%.3f%%) - Accepted:%i(%.3f%%) - biasedStdErr:%.6f\n" %(self.__generated , self.__tried, triedRatio, self.__accepted, acceptedRatio, self.biasedStdErr)) #<--
            frame = [self.allNames[idx]+ " " + "%10.5f"%self.realCoordinates[idx][0] + " %10.5f"%self.realCoordinates[idx][1] + " %10.5f"%self.realCoordinates[idx][2] + "\n" for idx in groupAtomsIndexes] #<--
            frame.extend([self.allNames[idx]+ " " + "%10.5f"%_coordsBeforeMove[idx][0] + " %10.5f"%_coordsBeforeMove[idx][1] + " %10.5f"%_coordsBeforeMove[idx][2] + "\n" for idx in range(_coordsBeforeMove.shape[0])]) #<--
            _xyzfd.write("".join(frame)) #<--
        #   #####################################################################################   #
        #   ################################# FINISH ENGINE RUN #################################   #        
        #-->LOGGER.info("Engine finishes executing all '%i' steps in %s" % (_numberOfSteps, get_elapsed_time(_engineStartTime, format="%d(days) %d:%d:%d")))
        # close .xyz file
        #-->if _xyzFrequency is not None:
        #-->    _xyzfd.close()
        _xyzfd.close() #<--
  
  
##########################################################################################
#####################################  CREATE ENGINE  ####################################  
# engine variables
pdbPath = "system.pdb"
# initialize engine
ENGINE = BiasedEngine(path=None)
ENGINE.set_pdb(pdbPath)
# create constraints
EMD_CONSTRAINT = InterMolecularDistanceConstraint()
# add constraints to engine
ENGINE.add_constraints([EMD_CONSTRAINT])


##########################################################################################
#####################################  DIFFERENT RUNS  ###################################
def refine(ENGINE, nsteps=1000, rang=5):
    # create group of single molecule
    ENGINE.set_groups_as_molecules()
    ENGINE.set_groups(ENGINE.groups[0])
    # set move generator
    [g.set_move_generator( MoveGeneratorCollector(collection=[TranslationGenerator(amplitude=2),RotationGenerator(amplitude=45)],randomize=True) ) for g in ENGINE.groups]
    # set selector
    recur = nsteps
    gs = RecursiveGroupSelector(RandomSelector(ENGINE), recur=recur, refine=True, explore=False)
    # run
    for step in range(rang):
        ENGINE.run(numberOfSteps=nsteps, saveFrequency=2*nsteps)  


        
##########################################################################################
#####################################  RUN SIMULATION  ###################################
# remove all .xyz trajectory files
files = [f for f in os.listdir(".") if os.path.isfile(f) and ".xyz" in f]
[os.remove(fname) for fname in files]
# run engine
refine(ENGINE, 1000)


##########################################################################################
##################################  VISUALIZE SIMULATION  ################################
ENGINE.set_pdb(pdbPath)
ENGINE.visualize( boxToCenter=True, boxWidth=1, 
                  otherParams = ['mol new trajectory.xyz', 'mol modstyle 0 1 VDW'],
                  representationParams='CPK 1.0 0.2 50 50')  




# standard library imports
from __future__ import print_function
import os, sys

# external libraries imports

# fullrmc imports
from fullrmc.Globals import FLOAT_TYPE, LOGGER
from fullrmc.Engine import Engine
from fullrmc.Core.Constraint import ExperimentalConstraint
from fullrmc.Constraints.PairDistributionConstraints import PairDistributionConstraint
from fullrmc.Constraints.DistanceConstraints import InterMolecularDistanceConstraint
from fullrmc.Constraints.AtomicCoordinationConstraints import AtomicCoordinationNumberConstraint
from fullrmc.Selectors.RandomSelectors import RandomSelector
from fullrmc.Core.GroupSelector import RecursiveGroupSelector
from fullrmc.Selectors.OrderedSelectors import DirectionalOrderSelector
from fullrmc.Core.Group import EmptyGroup
from fullrmc import MultiframeUtils

##########################################################################################
#####################################  CREATE ENGINE  ####################################
# dirname
try:
    DIR_PATH = os.path.dirname( os.path.realpath(__file__) )
except:
    DIR_PATH = ''


# files name
grFileName     = "SiOx.gr"
pdbFileName    = "SiOx.pdb"
engineFileName = "SiOx.rmc"
multiframe     = 'size_distribution'
numberOfFrames = 10
FRESH_START    = False
# engine variables
grExpPath      = os.path.join(DIR_PATH, grFileName)
pdbPath        = os.path.join(DIR_PATH, pdbFileName)
engineFilePath = os.path.join(DIR_PATH, engineFileName)


ENGINE = Engine(path=None)
if not ENGINE.is_engine(engineFilePath) or FRESH_START:
   # create engine
    ENGINE = Engine(path=engineFilePath, freshStart=True)
    ENGINE.set_pdb(pdbPath)
    # create and add pair distribution constraint to the engine
    PDF_CONSTRAINT = PairDistributionConstraint(experimentalData=grExpPath, weighting="atomicNumber")
    # shape function parameters
    params = {'rmin':0., 'rmax':None, 'dr':0.5,
              'qmin':0.0001, 'qmax':0.6, 'dq':0.005,
              'updateFreq':1000}
    PDF_CONSTRAINT.set_shape_function_parameters(params)
    ENGINE.add_constraints([PDF_CONSTRAINT])
    # Intermolecular constraint
    EMD_CONSTRAINT = InterMolecularDistanceConstraint()
    ENGINE.add_constraints([EMD_CONSTRAINT])
    EMD_CONSTRAINT.set_type_definition("element")
    EMD_CONSTRAINT.set_pairs_distance([('Si','Si',1.75), ('O','O',1.10), ('Si','O',1.30)])
    # coordination number constraint
    ACNC_CONSTRAINT = AtomicCoordinationNumberConstraint()
    ENGINE.add_constraints([ACNC_CONSTRAINT])
    ACNC_CONSTRAINT.set_coordination_number_definition( coordNumDef=[('Si','Si',1.8,2.8,3,6), ])
    # initialize constraints
    _ = ENGINE.initialize_used_constraints()
    # set number density
    ENGINE.set_number_density(.0125)
    # save engine
    ENGINE.save()
else:
    ENGINE = ENGINE.load(engineFilePath)
    # get constraints
    PDF_CONSTRAINT, EMD_CONSTRAINT, ACNC_CONSTRAINT = ENGINE.constraints



# add multiframe
if not ENGINE.is_frame(multiframe):
    ENGINE.add_frame({'name':multiframe, 'frames_name':numberOfFrames})
    for sf in ENGINE.frames[multiframe]['frames_name']:
        ENGINE.set_used_frame(os.path.join(multiframe, sf))
        ENGINE.set_pdb(os.path.join(DIR_PATH, 'multiframe_structure_%s.pdb'%sf))
        _usedConstraints, _constraints, _rigidConstraints = ENGINE.initialize_used_constraints(sortConstraints=True)
        if not len(_usedConstraints):
            LOGGER.warn("@%s No constraints are used. Configuration will be randomized"%ENGINE.usedFrame)
        # runtime initialize group selector
        ENGINE._Engine__groupSelector._runtime_initialize()
        # runtime initialize constraints
        _=[c._runtime_initialize() for c in _usedConstraints]
        LOGGER.info("@%s Stochastic engine files are created"%ENGINE.usedFrame)


################################################################################
###################### CREATE SOFTGRID WORKERS MANAGEMENT ######################
WM = MultiframeUtils.WorkersManagement()
WM.start(engine=ENGINE, multiframe='size_distribution', orchestrator=None)


################################################################################
######################## OPTIMIZE OXYGEN ATOMS POSITION ########################
for sf in ENGINE.frames[multiframe]['frames_name']:
    fname = os.path.join(multiframe, sf)
    ENGINE.set_used_frame(fname)
    # set groups as oxygen atoms only
    elements = ENGINE.get_original_data('allElements')
    groups   = [[idx] for idx, el in enumerate(elements) if el=='O']
    ENGINE.set_groups(groups)
    LOGGER.info("@%s Setting oxygen atoms groups (%i)"%(fname, len(groups),))

# run all frames independantly to optimize and fix oxygen atom positions
WM.run_independant(nCycle=10, numberOfSteps=10000, saveFrequency=1, cycleTimeout=3600)




################################################################################
######################### OPTIMIZE ALL ATOMS POSITION ##########################
for sf in ENGINE.frames[multiframe]['frames_name']:
    fname = os.path.join(multiframe, sf)
    ENGINE.set_used_frame(fname)
    # set groups as oxygen atoms only
    ENGINE.set_groups_as_atoms()
    LOGGER.info("@%s Setting groups as atoms (%i)"%(fname, len(ENGINE.groups)))

# run all frames independantly to optimize and fix oxygen atom positions
WM.run_independant(nCycle=50, numberOfSteps=10000, saveFrequency=1, cycleTimeout=3600)


################################################################################
################## MOVE ALL ATOMS IN ALL FRAMES SYNCHRONOUSLY ##################
for sf in ENGINE.frames[multiframe]['frames_name']:
    fname = os.path.join(multiframe, sf)
    ENGINE.set_used_frame(fname)
    # set groups as all atoms
    ENGINE.set_groups_as_atoms()
    LOGGER.info("@%s Setting groups as atoms (%i)"%(fname, len(ENGINE.groups)))

# run all frames as a mixture system of all subframes
WM.run_dependant(nCycle=10000, firstNAccepted=1, numberOfSteps=100, subframesWeight=None, normalize=1, driftTolerance=1)
WM.run_dependant(nCycle=10000, firstNAccepted=1, numberOfSteps=100, subframesWeight=10,   normalize=1, driftTolerance=1)


################################################################################
################## MOVE ALL ATOMS IN ALL FRAMES SYNCHRONOUSLY ##################
RECUR = 200
for sf in ENGINE.frames[multiframe]['frames_name']:
    fname = os.path.join(multiframe, sf)
    ENGINE.set_used_frame(fname)
    # set groups as all atoms
    ENGINE.set_groups_as_atoms()
    # create recurring random selector to explore
    gs = RecursiveGroupSelector(RandomSelector(ENGINE), recur=RECUR, refine=False, explore=True)
    ENGINE.set_group_selector(gs)
    LOGGER.info("@%s Setting groups as atoms (%i)"%(fname, len(ENGINE.groups)))

# run all frames as a mixture system of all subframes
WM.run_dependant(nCycle=10000, firstNAccepted=1, numberOfSteps=RECUR, saveFrequency=10, cycleTimeout=900)



################################################################################
################################  CALL plot.py  ################################
os.system("%s %s"%(sys.executable, os.path.join(DIR_PATH, 'plot.py')))



"""
DistanceConstraints contains classes for all constraints related to
distances between atoms.

.. inheritance-diagram:: fullrmc.Constraints.DistanceConstraints
    :parts: 1
"""
# standard libraries imports
from __future__ import print_function
import itertools

# external libraries imports
import numpy as np

# fullrmc imports
from ..Globals import INT_TYPE, FLOAT_TYPE, LOGGER
from ..Globals import str, long, unicode, bytes, basestring, range, xrange, maxint
from ..Core.Collection import is_number, raise_if_collected, reset_if_collected_out_of_date
from ..Core.Collection import get_caller_frames
from ..Core.Constraint import Constraint, SingularConstraint, RigidConstraint
from ..Core.atomic_distances import multiple_atomic_distances_coords, full_atomic_distances_coords, pair_elements_stats

class _DistanceConstraint(RigidConstraint, SingularConstraint):
    """
    This is the design class for all distance related constraints.
    It contains all common methods and definitions but it's forbiden to be instantiated.
    """
    def __init__(self, defaultDistance, typeDefinition,
                       pairsDistanceDefinition, flexible, rejectProbability):
        # disable instantiation
        assert not self.__class__.__name__ == "_DistanceConstraint", LOGGER.error("Instantiating '_DistanceConstraint' is forbidden.")
        # initialize constraint
        RigidConstraint.__init__(self, rejectProbability=rejectProbability)
        # set atoms collector data keys
        self._atomsCollector.set_data_keys( ('typesIndex', 'allTypes') )
        # set defaultDistance
        self.set_default_distance(defaultDistance)
        # set flexible
        self.set_flexible(flexible)
        # set type definition
        self.__pairsDistanceDefinition = None
        self.set_type_definition(typeDefinition, pairsDistanceDefinition)
        # set computation cost
        self.set_computation_cost(4.0)
        # set frame data
        FRAME_DATA = [d for d in self.FRAME_DATA]
        FRAME_DATA.extend(['_DistanceConstraint__defaultDistance',
                           '_DistanceConstraint__pairsDistanceDefinition',
                           '_DistanceConstraint__pairsDistance',
                           '_DistanceConstraint__flexible',
                           '_DistanceConstraint__lowerLimitArray',
                           '_DistanceConstraint__upperLimitArray',
                           '_DistanceConstraint__typePairs',
                           '_DistanceConstraint__typePairsIndex',
                           '_DistanceConstraint__typeDefinition',
                           '_DistanceConstraint__types',
                           '_DistanceConstraint__allTypes',
                           '_DistanceConstraint__numberOfTypes',
                           '_DistanceConstraint__typesIndex',
                           '_DistanceConstraint__numberOfAtomsPerType',] )
        RUNTIME_DATA = [d for d in self.RUNTIME_DATA]
        RUNTIME_DATA.extend( ['_DistanceConstraint__typesIndex',
                              '_DistanceConstraint__allTypes',
                              '_DistanceConstraint__numberOfAtomsPerType'] )
        object.__setattr__(self, 'FRAME_DATA',   tuple(FRAME_DATA)   )
        object.__setattr__(self, 'RUNTIME_DATA', tuple(RUNTIME_DATA) )

    @property
    def defaultDistance(self):
        """ Default minimum distance. """
        return self.__defaultDistance

    @property
    def pairsDistanceDefinition(self):
        """ Pairs distance definition. """
        return self.__pairsDistanceDefinition

    @property
    def pairsDistance(self):
        """ Pairs distance dictionary. """
        return self.__pairsDistance

    @property
    def flexible(self):
        """ Flexible flag. """
        return self.__flexible

    @property
    def lowerLimitArray(self):
        """
        Lower limit array used in distances calculation.
        for InterMolecularDistanceConstraint it's always a numpy.zeros array
        """
        return self.__lowerLimitArray

    @property
    def upperLimitArray(self):
        """
        Upper limit array used in distances calculation.
        for InterMolecularDistanceConstraint it's the minimum distance
        allowed between pair of intermolecular atoms.
        """
        return self.__upperLimitArray

    @property
    def typePairs(self):
        """ Atom's type pairs sorted list."""
        return self.__typePairs

    @property
    def typeDefinition(self):
        """ Atom's type definition. """
        return self.__typeDefinition

    @property
    def types(self):
        """ Atom's type set. """
        return self.__types

    @property
    def allTypes(self):
        """ All atoms type. """
        return self.__allTypes

    @property
    def numberOfTypes(self):
        """ Number of defined atom types in the configuration. """
        return self.__numberOfTypes

    @property
    def typesIndex(self):
        """ Type indexes list. """
        return self.__typesIndex

    @property
    def typePairsIndex(self):
        """Numpy array look up for type pairs index."""
        return self.__typePairsIndex

    @property
    def numberOfAtomsPerType(self):
        """ Number of atoms per type dict. """
        return self.__numberOfAtomsPerType

    def _on_collector_reset(self):
        pass

    def listen(self, message, argument=None):
        """
        listen to any message sent from the Broadcaster.

        :Parameters:
            #. message (object): Any python object to send to constraint's
               listen method.
            #. argument (object): Any type of argument to pass to the listeners.
        """
        if message in("engine set", "update molecules indexes"):
            self.set_type_definition(self.__typeDefinition, self.__pairsDistanceDefinition)
            # reset constraint is called in set_paris_distance
        elif message in("update boundary conditions",):
            self.reset_constraint()

    def set_flexible(self, flexible):
        """
        Set flexible flag.

        :Parameters:
            #. flexible (boolean): Whether to allow atoms to break constraints
               definition under the condition of decreasing total
               standardError. If flexible is set to False, atoms will never be
               allowed to cross from above to below minimum allowed distance.
               Even if the later will decrease some other unsatisfying atoms
               distances, and therefore the total standardError of the
               constraint.
        """
        assert isinstance(flexible, bool), LOGGER.error("flexible must be boolean")
        self.__flexible = flexible
        # dump to repository
        self._dump_to_repository({'_DistanceConstraint__flexible' :self.__flexible})

    def set_default_distance(self, defaultDistance):
        """
        Sets the default intermolecular minimum distance.

        :Parameters:
            #. defaultDistance (number): The default minimum distance.
        """
        assert is_number(defaultDistance), LOGGER.error("defaultDistance must be a number")
        defaultDistance = FLOAT_TYPE(defaultDistance)
        assert defaultDistance>=0, LOGGER.error("defaultDistance must be positive")
        self.__defaultDistance = defaultDistance
        # dump to repository
        self._dump_to_repository({'_DistanceConstraint__defaultDistance' :self.__defaultDistance})

    def set_type_definition(self, typeDefinition, pairsDistanceDefinition=None):
        """
        Alias to set_pairs_distance used when typeDefinition needs to
        be re-defined.

        :Parameters:
            #. typeDefinition (string): Can be either 'element' or 'name'.
               Sets the rules about how to differentiate between atoms and
               how to parse pairsLimits.
            #. pairsDistanceDefinition (None, list, set, tuple): The minimum
               distance set to every pair of elements. If None is given, the
               already defined pairsDistanceDefinition will be used and
               passed to set_pairs_distance method.
        """
        # set typeDefinition
        assert typeDefinition in ("name", "element"), LOGGER.error("typeDefinition must be either 'name' or 'element'")
        if self.engine is None:
            types                = None
            allTypes             = None
            numberOfTypes        = None
            typesIndex           = None
            numberOfAtomsPerType = None
        elif typeDefinition == "name":
            # copying because after loading and deserializing engine, pointer to
            # original data is lost and this will generate ambiguity in atoms collection
            types                = self.engine.get_original_data("names")
            allTypes             = self.engine.get_original_data("allNames")
            numberOfTypes        = len(types)
            typesIndex           = self.engine.get_original_data("namesIndex")
            numberOfAtomsPerType = self.engine.get_original_data("numberOfAtomsPerName")
        elif typeDefinition == "element":
            types                = self.engine.get_original_data("elements")
            allTypes             = self.engine.get_original_data("allElements")
            numberOfTypes        = len(types)
            typesIndex           = self.engine.get_original_data("elementsIndex")
            numberOfAtomsPerType = self.engine.get_original_data("numberOfAtomsPerElement")
        # set type definition
        self.__typeDefinition       = typeDefinition
        self.__types                = types
        self.__allTypes             = allTypes
        self.__numberOfTypes        = numberOfTypes
        self.__typesIndex           = typesIndex
        self.__numberOfAtomsPerType = numberOfAtomsPerType
        if self.__types is None:
            self.__typePairs        = None
            self.__typePairsIndex = [[],[]]
        else:
            self.__typePairs = sorted(itertools.combinations_with_replacement(self.__types,2))
            self.__typePairsIndex = [[],[]]
            for pair in self.__typePairs:
                idi = self.__types.index(pair[0])
                idj = self.__types.index(pair[1])
                self.__typePairsIndex[0].append(idi)
                self.__typePairsIndex[1].append(idj)
            self.__typePairsIndex = np.transpose(self.__typePairsIndex).astype(INT_TYPE)
        # dump to repository
        self._dump_to_repository({'_DistanceConstraint__typeDefinition'      :self.__typeDefinition,
                                  '_DistanceConstraint__types'               :self.__types,
                                  '_DistanceConstraint__allTypes'            :self.__allTypes,
                                  '_DistanceConstraint__numberOfTypes'       :self.__numberOfTypes,
                                  '_DistanceConstraint__typesIndex'          :self.__typesIndex,
                                  '_DistanceConstraint__numberOfAtomsPerType':self.__numberOfAtomsPerType,
                                  '_DistanceConstraint__typePairs'           :self.__typePairs,
                                  '_DistanceConstraint__typePairsIndex'      :self.__typePairsIndex})
        # set pair distance
        if pairsDistanceDefinition is None:
            pairsDistanceDefinition = self.__pairsDistanceDefinition
        self.set_pairs_distance(pairsDistanceDefinition)

    #@raise_if_collected
    def set_pairs_distance(self, pairsDistanceDefinition):
        """
        Set the pairs intermolecular minimum distance.

        :Parameters:
            #. pairsDistanceDefinition (None, list, set, tuple): The minimum
               distance to every pair of elements. A list of tuples must
               be given, all missing pairs will get automatically assigned the
               given default minimum distance value.
               First defined elements pair distance will cancel all redundant
               ones. If None is given, all pairs will be automatically
               generated and assigned the given defaultMinimumDistance value.

               ::

                   e.g. [('h','h',1.5), ('h','c',2.015), ...]

        """
        if self.engine is None:
            newPairsDistance = pairsDistanceDefinition
        elif pairsDistanceDefinition is None:
            newPairsDistance = {}
            for el1 in self.__types:
                newPairsDistance[el1] = {}
                for el2 in self.__types:
                    newPairsDistance[el1][el2] = self.__defaultDistance
        else:
            newPairsDistance = {}
            assert isinstance(pairsDistanceDefinition, (list, set, tuple)), LOGGER.error("pairsDistanceDefinition must be a list")
            for pair in pairsDistanceDefinition:
                assert isinstance(pair, (list, set, tuple)), LOGGER.error("pairsDistanceDefinition list items must be lists as well")
                pair = list(pair)
                assert len(pair)==3, LOGGER.error("pairsDistanceDefinition list pair item list must have three items")
                if pair[0] not in self.__types:
                    LOGGER.warn("pairsDistanceDefinition list pair item '%s' is not a valid engine type '%s', definition item omitted"%(pair[0], self.__typeDefinition) )
                    continue
                if pair[1] not in self.__types:
                    LOGGER.warn("pairsDistanceDefinition list pair item '%s' is not a valid engine type '%s', definition item omitted"%(pair[1], self.__typeDefinition) )
                    continue
                # create elements keys
                if not pair[0] in newPairsDistance:
                    newPairsDistance[pair[0]] = {}
                if not pair[1] in newPairsDistance:
                    newPairsDistance[pair[1]] = {}
                assert is_number(pair[2]), LOGGER.error("pairsDistanceDefinition list pair item list third item must be a number")
                distance = FLOAT_TYPE(pair[2])
                assert distance>=0, LOGGER.error("pairsDistanceDefinition list pair item list third item must be bigger than 0")
                # set minimum distance
                if pair[1] in newPairsDistance[pair[0]]:
                    LOGGER.warn("types pair ('%s','%s') distance definition is redundant, '%s' is omitted"%(pair[0], pair[1], pair))
                else:
                    newPairsDistance[pair[0]][pair[1]] = distance
                if pair[0] in newPairsDistance[pair[1]] and pair[0]!=pair[1]:
                    LOGGER.warn("types pair ('%s','%s') distance definition is redundant, '%s' is omitted"%(pair[1], pair[0], pair))
                else:
                    newPairsDistance[pair[1]][pair[0]] = distance
            # complete not defined distances
            for el1 in self.__types:
                if not el1 in newPairsDistance:
                    newPairsDistance[el1] = {}
                for el2 in self.__types:
                    if not el2 in newPairsDistance:
                        newPairsDistance[el2] = {}
                    if not el2 in newPairsDistance[el1]:
                        if el1 in newPairsDistance[el2]:
                            newPairsDistance[el1][el2] = newPairsDistance[el2][el1]
                        else:
                            LOGGER.warn("types pair ('%s','%s') distance definition is not defined and therefore it is set to the default distance '%s'"%(el1, el2, self.__defaultDistance))
                            newPairsDistance[el1][el2] = self.__defaultDistance
                            newPairsDistance[el2][el1] = self.__defaultDistance
                    assert newPairsDistance[el1][el2] == newPairsDistance[el2][el1], LOGGER.error("types '%s', and '%s' pair distance definitions are in conflict. (%s,%s, %s) and (%s,%s, %s)"%(el1,el2, el1,el2,newPairsDistance[el1][el2], el2,el1, newPairsDistance[el2][el1]))
        # set new pairsDistance value
        self.__pairsDistanceDefinition = pairsDistanceDefinition
        self.__pairsDistance           = newPairsDistance
        #if self.__pairsDistance is not None:
        if self.engine is not None:
            self.__lowerLimitArray = np.zeros((self.__numberOfTypes, self.__numberOfTypes, 1), dtype=FLOAT_TYPE)
            self.__upperLimitArray = np.zeros((self.__numberOfTypes, self.__numberOfTypes, 1), dtype=FLOAT_TYPE)
            for idx1 in range(self.__numberOfTypes):
                el1 = self.__types[idx1]
                for idx2 in range(self.__numberOfTypes):
                    el2  = self.__types[idx2]
                    dist = self.__pairsDistance[el1][el2]
                    self.__upperLimitArray[idx1,idx2,0] = FLOAT_TYPE(dist)
                    self.__upperLimitArray[idx2,idx1,0] = FLOAT_TYPE(dist)
        else:
            self.__lowerLimitArray = None
            self.__upperLimitArray = None
        # dump to repository
        self._dump_to_repository({'_DistanceConstraint__pairsDistanceDefinition': self.__pairsDistanceDefinition,
                                  '_DistanceConstraint__pairsDistance'          : self.__pairsDistance,
                                  '_DistanceConstraint__lowerLimitArray'        : self.__lowerLimitArray,
                                  '_DistanceConstraint__upperLimitArray'        : self.__upperLimitArray})
        # reset constraint
        self.reset_constraint() # ADDED 2017-JAN-08

    def should_step_get_rejected(self, standardError):
        """
        Given a standardError, return whether to keep or reject new
        standardError according to the constraint rejectProbability.
        In addition, if flexible flag is set to True, total number of atoms
        not satisfying constraints definition must be decreasing or at least
        remain the same.

        :Parameters:
            #. standardError (number): Standard error to compare with
               Constraint's standard error.

        :Returns:
            #. result (boolean): True to reject step, False to accept.
        """
        if self.__flexible:
            # compute if step should get rejected as a RigidConstraint
            return super(_DistanceConstraint, self).should_step_get_rejected(standardError)
        else:
            cond = self.activeAtomsDataAfterMove["number"]>self.activeAtomsDataBeforeMove["number"]
            if np.any(cond):
                return True
            return False

    def _compute_standard_error(self, distances):
        return FLOAT_TYPE( np.sum(distances) )

    def compute_standard_error(self, data):
        """
        Compute the standard error (stdErr) of data not satisfying constraint's
        conditions.

        .. math::
            stdErr = \\sum \\limits_{i}^{N} \\sum \\limits_{i+1}^{N}
            \\left| d_{ij}-D_{ij}) \\right|
            \\int_{0}^{D_{ij}} \\delta(x-d_{ij}) dx

        Where:\n
        :math:`N` is the total number of atoms in the system. \n
        :math:`D_{ij}` is the distance constraint set for atoms pair (i,j). \n
        :math:`d_{ij}` is the distance between atom i and atom j. \n
        :math:`\\delta` is the Dirac delta function. \n
        :math:`\\int_{0}^{D_{ij}} \\delta(x-d_{ij}) dx`
        is equal to 1 if :math:`0 \\leqslant d_{ij} \\leqslant D_{ij}` and 0 elsewhere.\n

        :Parameters:
            #. data (dict): data used to compute standard error.

        :Returns:
            #. standardError (number): The calculated standardError.
        """
        standardError = np.sum(data.values())
        return FLOAT_TYPE(standardError)

    def _get_constraint_value(self):
        if self.data is None:
            LOGGER.warn("data must be computed first using 'compute_data' method.")
            return np.array([], dtype=FLOAT_TYPE)
        # compute distances
        idi = self.__typePairsIndex[:,0]
        idj = self.__typePairsIndex[:,1]
        numbers   = (self.data["number"][idi,idj] + self.data["number"][idj,idi]).reshape(-1)
        distances = (self.data["distanceSum"][idi,idj] + self.data["distanceSum"][idj,idi]).reshape(-1)
        nonZero   = np.where(numbers)
        distances[nonZero] /= numbers[nonZero]
        return distances

    def get_constraint_value(self):
        """
        Get constraint's formatted dictionary data.

        :Returns:
            #. data (dictionary): Formatted dictionary data. Keys are
               type pairs and values constraint data.
        """
        if self.data is None:
            LOGGER.warn("data must be computed first using 'compute_data' method.")
            return {}
        # compute distances
        distances = self._get_constraint_value()
        # compute output
        output = {}
        for idx, pair in enumerate(self.__typePairs):
            key = "md_%s-%s"% pair
            output[key] = distances[idx]
        # return
        return output

    def _on_collector_collect_atom(self, realIndex):
        # get relative index
        relativeIndex = self._atomsCollector.get_relative_index(realIndex)
        # create dataDict
        dataDict = {}
        dataDict['typesIndex'] = self.typesIndex[relativeIndex]
        dataDict['allTypes']    = self.allTypes[relativeIndex]
        # reduce all indexes above relativeIndex in typesIndex
        # delete data
        self.__typesIndex = np.delete(self.__typesIndex, relativeIndex, axis=0)
        self.__allTypes    = np.delete(self.__allTypes,     relativeIndex, axis=0)
        self.__numberOfAtomsPerType[dataDict['allTypes']] -= 1
        # collect atom
        self._atomsCollector.collect(realIndex, dataDict=dataDict)

    def _on_collector_release_atom(self, realIndex):
        pass


class _MolecularDistanceConstraint(_DistanceConstraint):
    """
    This is the design class for all molecular distance related constraints.
    It contains all common methods and definitions but it's forbiden to be instantiated.
    """
    def __init__(self, *args, **kwargs):
        assert not self.__class__.__name__ == "_MolecularDistanceConstraint", LOGGER.error("Instantiating '_MolecularDistanceConstraint' is forbidden.")
        # check customizable flags
        assert self._interMolecular or self._intraMolecular, LOGGER.error("In _MolecularDistanceConstraint either '_interMolecular' or '_intraMolecular' must be set to True")
        assert not (self._interMolecular and self._intraMolecular), LOGGER.error("In _MolecularDistanceConstraint both '_interMolecular' and '_intraMolecular' can't be set to True")
        # initialize constraint
        super(_MolecularDistanceConstraint, self).__init__(*args, **kwargs)
        # creating fixed flags
        object.__setattr__(self, '_countWithinLimits', True)
        object.__setattr__(self, '_reduceDistance', False)
        object.__setattr__(self, '_reduceDistanceToUpper', True)
        object.__setattr__(self, '_reduceDistanceToLower', False)
        # set frame data
        FRAME_DATA = [d for d in self.FRAME_DATA]
        FRAME_DATA.extend(['_MolecularDistanceConstraint__typesStats', ] )
        object.__setattr__(self, 'FRAME_DATA',   tuple(FRAME_DATA)   )


    def __setattr__(self, name, value):
        if name in ('_interMolecular','_intraMolecular','_countWithinLimits','_reduceDistance','_reduceDistanceToUpper','_reduceDistanceToLower'):
            raise Exception( LOGGER.error("setting '%'s is forbiden"%name) )
        object.__setattr__(self, name, value)


    def set_type_definition(self, typeDefinition, pairsDistanceDefinition=None):
        """
        Alias to set_pairs_distance used when typeDefinition needs
        to be re-defined.

        :Parameters:
            #. typeDefinition (string): Can be either 'element' or 'name'.
               Sets the rules about how to differentiate between atoms and
               how to parse pairsLimits.
            #. pairsDistanceDefinition (None, list, set, tuple): Minimum
               distance to every pair of elements. If None is given, the
               already defined pairsDistanceDefinition will be used and
               passed to set_pairs_distance method.
        """
        super(_MolecularDistanceConstraint, self).set_type_definition(typeDefinition          = typeDefinition,
                                                                      pairsDistanceDefinition = pairsDistanceDefinition)
        # create stats
        stats = None
        if self.types is None:
            stats = None
        elif self._interMolecular:
            _, stats = pair_elements_stats(elementIndex     = self.typesIndex,
                                           numberOfElements = self.numberOfTypes,
                                           moleculeIndex    = self.engine.moleculesIndex)
        else:
            stats, _ = pair_elements_stats(elementIndex     = self.typesIndex,
                                           numberOfElements = self.numberOfTypes,
                                           moleculeIndex    = self.engine.moleculesIndex)
        # set types stats
        self.__typesStats = stats
        # dump to repository
        self._dump_to_repository({'_MolecularDistanceConstraint__typesStats' :self.__typesStats})

    @property
    def typesStats(self):
        """Numpy array of number of found pairs in system."""
        return self.__typesStats

    @reset_if_collected_out_of_date
    def compute_data(self):
        """ Compute constraint's data."""
        # compute data
        nintra,dintra, ninter,dinter = \
        full_atomic_distances_coords( boxCoords             = self.engine.boxCoordinates,
                                      basis                 = self.engine.basisVectors,
                                      isPBC                 = self.engine.isPBC,
                                      moleculeIndex         = self.engine.moleculesIndex,
                                      elementIndex          = self.typesIndex,
                                      numberOfElements      = self.numberOfTypes,
                                      lowerLimit            = self.lowerLimitArray,
                                      upperLimit            = self.upperLimitArray,
                                      interMolecular        = self._interMolecular,
                                      intraMolecular        = self._intraMolecular,
                                      reduceDistance        = self._reduceDistance,
                                      reduceDistanceToUpper = self._reduceDistanceToUpper,
                                      reduceDistanceToLower = self._reduceDistanceToLower,
                                      countWithinLimits     = self._countWithinLimits,
                                      ncores                = self.engine._runtime_ncores)
        if self._interMolecular:
            number      = ninter
            distanceSum = dinter
        else:
            number      = nintra
            distanceSum = dintra
        # update data
        self.set_data( {"number":number, "distanceSum":distanceSum} )
        self.set_active_atoms_data_before_move(None)
        self.set_active_atoms_data_after_move(None)
        # set standardError
        self.set_standard_error( self._compute_standard_error(distances = self._get_constraint_value()) )
        # set original data
        if self.originalData is None:
            self._set_original_data(self.data)

    def compute_before_move(self, realIndexes, relativeIndexes):
        """
        Compute constraint before move is executed

        :Parameters:
            #. realIndexes (numpy.ndarray): Not used here.
            #. relativeIndexes (numpy.ndarray): Group atoms relative index
               the move will be applied to.
        """
        nintraM,dintraM, ninterM,dinterM = \
        multiple_atomic_distances_coords( indexes               = relativeIndexes,
                                          boxCoords             = self.engine.boxCoordinates,
                                          basis                 = self.engine.basisVectors,
                                          isPBC                 = self.engine.isPBC,
                                          moleculeIndex         = self.engine.moleculesIndex,
                                          elementIndex          = self.typesIndex,
                                          numberOfElements      = self.numberOfTypes,
                                          lowerLimit            = self.lowerLimitArray,
                                          upperLimit            = self.upperLimitArray,
                                          allAtoms              = True,
                                          countWithinLimits     = self._countWithinLimits,
                                          reduceDistance        = self._reduceDistance,
                                          reduceDistanceToUpper = self._reduceDistanceToUpper,
                                          reduceDistanceToLower = self._reduceDistanceToLower,
                                          interMolecular        = self._interMolecular,
                                          intraMolecular        = self._intraMolecular,
                                          ncores                = self.engine._runtime_ncores)
        nintraF,dintraF, ninterF,dinterF = \
        full_atomic_distances_coords( boxCoords             = self.engine.boxCoordinates[relativeIndexes],
                                      basis                 = self.engine.basisVectors,
                                      isPBC                 = self.engine.isPBC,
                                      moleculeIndex         = self.engine.moleculesIndex[relativeIndexes],
                                      elementIndex          = self.typesIndex[relativeIndexes],
                                      numberOfElements      = self.numberOfTypes,
                                      lowerLimit            = self.lowerLimitArray,
                                      upperLimit            = self.upperLimitArray,
                                      interMolecular        = self._interMolecular,
                                      intraMolecular        = self._intraMolecular,
                                      reduceDistance        = self._reduceDistance,
                                      reduceDistanceToUpper = self._reduceDistanceToUpper,
                                      reduceDistanceToLower = self._reduceDistanceToLower,
                                      countWithinLimits     = self._countWithinLimits,
                                      ncores                = self.engine._runtime_ncores)
        # check if inter or intra
        if self._interMolecular:
            numberM      = ninterM
            distanceSumM = dinterM
            numberF      = ninterF
            distanceSumF = dinterF
        else:
            numberM      = nintraM
            distanceSumM = dintraM
            numberF      = nintraF
            distanceSumF = dintraF
        # set active atoms data
        self.set_active_atoms_data_before_move( {"number":numberM-numberF, "distanceSum":distanceSumM-distanceSumF} )
        self.set_active_atoms_data_after_move(None)

    def compute_after_move(self, realIndexes, relativeIndexes, movedBoxCoordinates):
        """
        Compute constraint's data after move is executed.

        :Parameters:
            #. realIndexes (numpy.ndarray): Not used here.
            #. relativeIndexes (numpy.ndarray): Group atoms relative index
               the move will be applied to.
            #. movedBoxCoordinates (numpy.ndarray): The moved atoms new coordinates.
        """
        # change coordinates temporarily
        boxData = np.array(self.engine.boxCoordinates[relativeIndexes], dtype=FLOAT_TYPE)
        self.engine.boxCoordinates[relativeIndexes] = movedBoxCoordinates
        # calculate pair distribution function
        nintraM,dintraM, ninterM,dinterM = \
        multiple_atomic_distances_coords( indexes               = relativeIndexes,
                                          boxCoords             = self.engine.boxCoordinates,
                                          basis                 = self.engine.basisVectors,
                                          isPBC                 = self.engine.isPBC,
                                          moleculeIndex         = self.engine.moleculesIndex,
                                          elementIndex          = self.typesIndex,
                                          numberOfElements      = self.numberOfTypes,
                                          lowerLimit            = self.lowerLimitArray,
                                          upperLimit            = self.upperLimitArray,
                                          allAtoms              = True,
                                          countWithinLimits     = self._countWithinLimits,
                                          reduceDistance        = self._reduceDistance,
                                          reduceDistanceToUpper = self._reduceDistanceToUpper,
                                          reduceDistanceToLower = self._reduceDistanceToLower,
                                          interMolecular        = self._interMolecular,
                                          intraMolecular        = self._intraMolecular,
                                          ncores                = self.engine._runtime_ncores)
        nintraF,dintraF, ninterF,dinterF = \
        full_atomic_distances_coords( boxCoords             = self.engine.boxCoordinates[relativeIndexes],
                                      basis                 = self.engine.basisVectors,
                                      isPBC                 = self.engine.isPBC,
                                      moleculeIndex         = self.engine.moleculesIndex[relativeIndexes],
                                      elementIndex          = self.typesIndex[relativeIndexes],
                                      numberOfElements      = self.numberOfTypes,
                                      lowerLimit            = self.lowerLimitArray,
                                      upperLimit            = self.upperLimitArray,
                                      interMolecular        = self._interMolecular,
                                      intraMolecular        = self._intraMolecular,
                                      reduceDistance        = self._reduceDistance,
                                      reduceDistanceToUpper = self._reduceDistanceToUpper,
                                      reduceDistanceToLower = self._reduceDistanceToLower,
                                      countWithinLimits     = self._countWithinLimits,
                                      ncores                = self.engine._runtime_ncores)
        # check if inter or intra
        if self._interMolecular:
            numberM      = ninterM
            distanceSumM = dinterM
            numberF      = ninterF
            distanceSumF = dinterF
        else:
            numberM      = nintraM
            distanceSumM = dintraM
            numberF      = nintraF
            distanceSumF = dintraF
        # set active atoms data
        self.set_active_atoms_data_after_move( {"number":numberM-numberF, "distanceSum":distanceSumM-distanceSumF} )
        # reset coordinates
        self.engine.boxCoordinates[relativeIndexes] = boxData
        # compute standardError after move
        number = self.data["number"]-self.activeAtomsDataBeforeMove["number"]+self.activeAtomsDataAfterMove["number"]
        distanceSum = self.data["distanceSum"]-self.activeAtomsDataBeforeMove["distanceSum"]+self.activeAtomsDataAfterMove["distanceSum"]
        data = self.data
        # change temporarily data attribute
        self.set_data( {"number":number, "distanceSum":distanceSum} )
        self.set_after_move_standard_error( self._compute_standard_error(distances = self._get_constraint_value()) )
        # change back data attribute
        self.set_data( data )

    def accept_move(self, realIndexes, relativeIndexes):
        """
        Accept move.

        :Parameters:
            #. realIndexes (numpy.ndarray): Not used here.
            #. relativeIndexes (numpy.ndarray): Not used here.
        """
        number = self.data["number"]-self.activeAtomsDataBeforeMove["number"]+self.activeAtomsDataAfterMove["number"]
        distanceSum = self.data["distanceSum"]-self.activeAtomsDataBeforeMove["distanceSum"]+self.activeAtomsDataAfterMove["distanceSum"]
        # change permanently data attribute
        self.set_data( {"number":number, "distanceSum":distanceSum} )
        # reset activeAtoms data
        self.set_active_atoms_data_before_move(None)
        self.set_active_atoms_data_after_move(None)
        # update standardError
        self.set_standard_error( self.afterMoveStandardError )
        self.set_after_move_standard_error( None )

    def reject_move(self, realIndexes, relativeIndexes):
        """
        Reject move.

        :Parameters:
            #. realIndexes (numpy.ndarray): Not used here.
            #. relativeIndexes (numpy.ndarray): Not used here.
        """
        # reset activeAtoms data
        self.set_active_atoms_data_before_move(None)
        self.set_active_atoms_data_after_move(None)
        # update standardError
        self.set_after_move_standard_error( None )

    def compute_as_if_amputated(self, realIndex, relativeIndex):
        """
        Compute and return constraint's data and standard error as if atom given its
        its was amputated.

        :Parameters:
            #. realIndex (numpy.ndarray): Not used here.
            #. relativeIndex (numpy.ndarray): Not used here.
        """
        pass

    def accept_amputation(self, realIndex, relativeIndex):
        """
        Accept amputation of atom and sets constraints data and standard error accordingly.

        :Parameters:
            #. realIndex (numpy.ndarray): Atom's index as a numpy array
               of a single element.
            #. relativeIndex (numpy.ndarray): Atom's relative index as a
               numpy array of a single element.
        """
        # MAYBE WE DON"T NEED TO CHANGE DATA AND SE. BECAUSE THIS MIGHT BE A PROBLEM
        # WHEN IMPLEMENTING ATOMS RELEASING. MAYBE WE NEED TO COLLECT DATA INSTEAD, REMOVE
        # AND ADD UPON RELEASE
        self.compute_before_move(realIndexes=realIndex, relativeIndexes=relativeIndex)
        #self.compute_before_move(indexes = np.array([index], dtype=INT_TYPE) )
        # change permanently data attribute
        number      = self.data["number"]-self.activeAtomsDataBeforeMove["number"]
        distanceSum = self.data["distanceSum"]-self.activeAtomsDataBeforeMove["distanceSum"]
        self.set_data( {"number":number, "distanceSum":distanceSum} )
        # reset activeAtoms data
        self.set_active_atoms_data_before_move(None)
        # update standardError
        SE = self._compute_standard_error(distances = self._get_constraint_value())
        self.set_standard_error( SE )

    def reject_amputation(self, realIndex, relativeIndex):
        """
        Reject amputation of atom.

        :Parameters:
            #. realIndex (numpy.ndarray): No used here.
            #. relativeIndex (numpy.ndarray): Not used here.
        """
        pass

    def plot(self, ax=None, width=0.6,
                   correctColor = '#0066ff',
                   erroneousColor = '#d62728',
                   xlabel=True, xlabelSize=16,
                   ylabel=True, ylabelSize=16,
                   legend=True, legendCols=1, legendLoc='best', legendText=('correct','erroneous'),
                   title=True, titleStdErr=True, titleAtRem=True,
                   titleUsedFrame=True, show=True):
        """
        Plot distance constraint's data.

        :Parameters:
            #. ax (None, matplotlib Axes): matplotlib Axes instance to plot in.
               If None is given a new plot figure will be created.
            #. width (number): Bars width, must be >0 and <=1
            #. correctColor (color): correct data bars color.
            #. erroneousColor (color): erroneous data bars color.
            #. xlabel (boolean): Whether to create x label.
            #. xlabelSize (number): The x label font size.
            #. ylabel (boolean): Whether to create y label.
            #. ylabelSize (number): The y label font size.
            #. legend (boolean): Whether to create the legend or not
            #. legendCols (integer): Legend number of columns.
            #. legendLoc (string): The legend location. Anything among
               'right', 'center left', 'upper right', 'lower right', 'best',
               'center', 'lower left', 'center right', 'upper left',
               'upper center', 'lower center' is accepted.
            #. legendText (list): The legend of number of correct pairs
               satisfying constraint and number of erroneous pairs unsatisfying
               constraint condition.
            #. title (boolean): Whether to create the title or not
            #. titleStdErr (boolean): Whether to show constraint standard error
               value in title.
            #. titleAtRem (boolean): Whether to show engine's number of removed
               atoms.
            #. titleUsedFrame(boolean): Whether to show used frame name in
               title.
            #. show (boolean): Whether to render and show figure before
               returning.

        :Returns:
            #. figure (matplotlib Figure): matplotlib used figure.
            #. axes (matplotlib Axes): matplotlib used axes.
        """
        # get constraint value
        if self.data is None:
            LOGGER.warn("%s constraint data are not computed."%(self.__class__.__name__))
            return
        # check width
        assert 0<width<=1, LOGGER.error("width must be a number between 0 and 1")
        # import matplotlib
        import matplotlib.pyplot as plt
        # get axes
        if ax is None:
            FIG  = plt.figure()
            AXES = plt.gca()
        else:
            AXES = ax
            FIG  = AXES.get_figure()
        # get numbers and differences
        idi = self.typePairsIndex[:,0]
        idj = self.typePairsIndex[:,1]
        numbers = (self.data["number"][idi,idj] + self.data["number"][idj,idi]).reshape(-1).astype(FLOAT_TYPE)
        stats   = (self.typesStats[idi,idj] + self.typesStats[idj,idi]).reshape(-1).astype(FLOAT_TYPE)
        diff    = stats-numbers
        # plot bars
        ind   = np.arange(1,len(numbers)+1)  # the x locations for the groups
        p1 = AXES.bar(ind, diff, width, color=correctColor, label=legendText[0])
        p2 = AXES.bar(ind, numbers, width, bottom=diff, color=erroneousColor, label=legendText[1])
        AXES.set_xlim(0,len(numbers)+1.5)
        AXES.set_ylim(0, max(stats)+0.15*max(stats))
        # set ticks
        plt.xticks(ind, ["%s-%s"%(el1,el2) for el1,el2 in self.typePairs])
        # ratio labels
        data = self._get_constraint_value()
        for d, rect in zip(data, AXES.patches):
            height = rect.get_height()
            t = AXES.text(x     = rect.get_x() + rect.get_width()/2,
                          y     = height + 5,
                          s     = " "+str(d),
                          color = 'black',
                          rotation = 90,
                          horizontalalignment = 'center',
                          verticalalignment   = 'bottom')
        # set axis labels
        if xlabel:
            AXES.set_xlabel("Type pairs", size=xlabelSize)
        if ylabel:
            AXES.set_ylabel("Number of pairs"  , size=ylabelSize)
        # set title
        if title:
            FIG.canvas.set_window_title('Molecular Distances Constraint')
            if titleUsedFrame:
                t = '$frame: %s$ : '%self.engine.usedFrame.replace('_','\\_')
            else:
                t = ''
            if titleAtRem:
                t += "$%i$ $rem.$ $at.$ - "%(len(self.engine._atomsCollector))
            if titleStdErr and self.standardError is not None:
                t += "$std$ $error=%.6f$ "%(self.standardError)
            if len(t):
                AXES.set_title(t)
        # set background color
        FIG.patch.set_facecolor('white')
        # plot legend
        if legend:
            AXES.legend(frameon=False, ncol=legendCols, loc=legendLoc)
        #show
        if show:
            plt.show()
        # return axes
        return FIG, AXES

    def export(self, fname, delimiter='     ', comments='# '):
        """
        Export distance constraint's data.

        :Parameters:
            #. fname (path): full file name and path.
            #. delimiter (string): String or character separating columns.
            #. comments (string): String that will be prepended to the header.
        """
        # get constraint value
        output = self.get_constraint_value()
        if not len(output):
            LOGGER.warn("%s constraint data are not computed."%(self.__class__.__name__))
            return
        # get numbers and differences
        idi = self.typePairsIndex[:,0]
        idj = self.typePairsIndex[:,1]
        numbers = (self.data["number"][idi,idj] + self.data["number"][idj,idi]).reshape(-1).astype(FLOAT_TYPE)
        stats   = (self.typesStats[idi,idj] + self.typesStats[idj,idi]).reshape(-1).astype(FLOAT_TYPE)
        diff    = stats-numbers
        # set data as strings
        stats   = [str(i) for i in stats]
        diff    = [str(i) for i in diff]
        numbers = [str(i) for i in numbers]
        pairs   = ["%s-%s"%(el1,el2) for el1,el2 in self.typePairs]
        stdErr  = self._get_constraint_value()
        # start creating header and data
        header = ["description","num_pairs","correct","erroneous","standard_error"]
        data   = [pairs,stats,diff,numbers,stdErr]
        # save
        data = np.transpose(data)
        np.savetxt(fname     = fname,
                   X         = data,
                   fmt       = '%s',
                   delimiter = delimiter,
                   header    = " ".join(header),
                   comments  = comments)


class InterMolecularDistanceConstraint(_MolecularDistanceConstraint):
    """
    Its controls the inter-molecular distances between atoms.

    +----------------------------------------------------------------------+
    |.. figure:: molecular_distances_constraint_plot_method.png            |
    |   :width: 530px                                                      |
    |   :height: 400px                                                     |
    |   :align: left                                                       |
    +----------------------------------------------------------------------+


    :Parameters:
        #. defaultDistance (number): The minimum distance allowed set by
           default for all atoms type.
        #. typeDefinition (string): Can be either 'element' or 'name'.
           Sets the rules about how to differentiate between atoms and how
           to parse pairsLimits.
        #. pairsDistanceDefinition (None, list, set, tuple): The minimum
           distance set to every pair of elements.
           A list of tuples must be given, all missing pairs will get
           automatically assigned the given defaultMinimumDistance value.
           First defined elements pair distance will cancel all redundant.
           If None is given all pairs will be automatically generated and
           assigned the given defaultMinimumDistance value.

           ::

               e.g. [('h','h',1.5), ('h','c',2.015), ...]

        #. flexible (boolean): Whether to allow atoms to break constraint's
           definition under the condition of decreasing total standardError
           of the constraint. If flexible is set to False, atoms will never
           be allowed to cross from above to below minimum allowed distance.
           Even if the later will decrease some other unsatisfying atoms
           distances, and therefore the total standardError of the constraint.
        #. rejectProbability (Number): rejecting probability of all steps where
           standardError increases. It must be between 0 and 1 where 1 means
           rejecting all steps where standardError increases and 0 means
           accepting all steps regardless whether standardError increases or
           not.


    .. code-block:: python

        # import fullrmc modules
        from fullrmc.Engine import Engine
        from fullrmc.Constraints.DistanceConstraints import InterMolecularDistanceConstraint

        # create engine
        ENGINE = Engine(path='my_engine.rmc')

        # set pdb file
        ENGINE.set_pdb('system.pdb')

        # create and add constraint
        EMD = InterMolecularDistanceConstraint()
        ENGINE.add_constraints(EMD)

        # create definition
        EMD.set_pairs_distance([('Si','Si',1.75), ('O','O',1.10), ('Si','O',1.30)])


    .. autoattribute:: defaultDistance
    .. autoattribute:: pairsDistanceDefinition
    .. autoattribute:: pairsDistance
    .. autoattribute:: flexible
    .. autoattribute:: lowerLimitArray
    .. autoattribute:: upperLimitArray
    .. autoattribute:: typePairs
    .. autoattribute:: typeDefinition
    .. autoattribute:: types
    .. autoattribute:: allTypes
    .. autoattribute:: numberOfTypes
    .. autoattribute:: typesIndex
    .. autoattribute:: numberOfAtomsPerType
    .. automethod:: listen
    .. automethod:: set_flexible
    .. automethod:: set_default_distance
    .. automethod:: set_type_definition
    .. automethod:: set_pairs_distance
    .. automethod:: should_step_get_rejected
    .. automethod:: compute_standard_error
    .. automethod:: get_constraint_value
    .. automethod:: compute_data
    .. automethod:: compute_before_move
    .. automethod:: compute_after_move
    .. automethod:: accept_move
    .. automethod:: reject_move
    .. automethod:: compute_as_if_amputated
    .. automethod:: accept_amputation
    .. automethod:: reject_amputation
    .. automethod:: plot
    .. automethod:: export
    """
    def __init__(self, defaultDistance=1.5, typeDefinition='element',
                       pairsDistanceDefinition=None, flexible=True, rejectProbability=1):
        # creating customizable flags
        object.__setattr__(self, '_interMolecular', True)
        object.__setattr__(self, '_intraMolecular', False)
        # initialize constraint
        super(InterMolecularDistanceConstraint, self).__init__(defaultDistance         = defaultDistance,
                                                               typeDefinition          = typeDefinition,
                                                               pairsDistanceDefinition = pairsDistanceDefinition,
                                                               flexible                = flexible,
                                                               rejectProbability       = rejectProbability)



class IntraMolecularDistanceConstraint(_MolecularDistanceConstraint):
    """
    .. py:class::IntraMolecularDistanceConstraint

    Its controls the intra-molecular distances between atoms.


    +----------------------------------------------------------------------+
    |.. figure:: molecular_distances_constraint_plot_method.png            |
    |   :width: 530px                                                      |
    |   :height: 400px                                                     |
    |   :align: left                                                       |
    +----------------------------------------------------------------------+


    :Parameters:
        #. defaultDistance (number): The minimum distance allowed set by
           default for all atoms type.
        #. typeDefinition (string): Can be either 'element' or 'name'.
           Sets the rules about how to differentiate between atoms and how to
           parse pairsLimits.
        #. pairsDistanceDefinition (None, list, set, tuple): The minimum
           distance set to every pair of elements. A list of tuples must be
           given, all missing pairs will get automatically assigned the given
           default minimum distance value. First defined elements pair distance
           will cancel all redundant. If None is given all pairs will be
           automatically generated and assigned the given default minimum
           distance value.

           ::

               e.g. [('h','h',1.5), ('h','c',2.015), ...]

        #. flexible (boolean): Whether to allow atoms to break constraint's
           definition under the condition of decreasing total standardError of
           the constraint. If flexible is set to False, atoms will never be
           allowed to cross from above to below minimum allowed distance.
           Even if the later will decrease some other unsatisfying atoms
           distances, and therefore the total standardError of the constraint.
        #. rejectProbability (Number): rejecting probability of all steps where
           standardError increases. It must be between 0 and 1 where 1 means
           rejecting all steps where standardError increases and 0 means
           accepting all steps regardless whether standardError increases or
           not.


    .. code-block:: python

        # import fullrmc modules
        from fullrmc.Engine import Engine
        from fullrmc.Constraints.DistanceConstraints import IntraMolecularDistanceConstraint

        # create engine
        ENGINE = Engine(path='my_engine.rmc')

        # set pdb file
        ENGINE.set_pdb('system.pdb')

        # create and add constraint
        EMD = IntraMolecularDistanceConstraint()
        ENGINE.add_constraints(EMD)

        # create definition
        EMD.set_pairs_distance([('Si','Si',1.75), ('O','O',1.10), ('Si','O',1.30)])


    .. autoattribute:: defaultDistance
    .. autoattribute:: pairsDistanceDefinition
    .. autoattribute:: pairsDistance
    .. autoattribute:: flexible
    .. autoattribute:: lowerLimitArray
    .. autoattribute:: upperLimitArray
    .. autoattribute:: typePairs
    .. autoattribute:: typeDefinition
    .. autoattribute:: types
    .. autoattribute:: allTypes
    .. autoattribute:: numberOfTypes
    .. autoattribute:: typesIndex
    .. autoattribute:: numberOfAtomsPerType
    .. automethod:: listen
    .. automethod:: set_flexible
    .. automethod:: set_default_distance
    .. automethod:: set_type_definition
    .. automethod:: set_pairs_distance
    .. automethod:: should_step_get_rejected
    .. automethod:: compute_standard_error
    .. automethod:: get_constraint_value
    .. automethod:: compute_data
    .. automethod:: compute_before_move
    .. automethod:: compute_after_move
    .. automethod:: accept_move
    .. automethod:: reject_move
    .. automethod:: compute_as_if_amputated
    .. automethod:: accept_amputation
    .. automethod:: reject_amputation
    .. automethod:: plot
    .. automethod:: export
    """
    def __init__(self, defaultDistance=1.5, typeDefinition='name',
                       pairsDistanceDefinition=None, flexible=True, rejectProbability=1):
        # creating customizable flags
        object.__setattr__(self, '_interMolecular', False)
        object.__setattr__(self, '_intraMolecular', True)
        # initialize constraint
        super(IntraMolecularDistanceConstraint, self).__init__(defaultDistance         = defaultDistance,
                                                               typeDefinition          = typeDefinition,
                                                               pairsDistanceDefinition = pairsDistanceDefinition,
                                                               flexible                = flexible,
                                                               rejectProbability       = rejectProbability)



"""
This is a C compiled module to compute pair distances histograms.
""" 
#from libc.math cimport sqrt, abs
import cython
cimport cython
import numpy as np
cimport numpy as np
from numpy cimport ndarray
from cython.parallel import prange
from fullrmc.Core.pairs_distances import pairs_distances_to_indexcoords


# declare types
NUMPY_FLOAT32 = np.float32
NUMPY_INT32   = np.int32
ctypedef np.float32_t C_FLOAT32
ctypedef np.int32_t   C_INT32

# declare constants
cdef C_FLOAT32 BOX_LENGTH      = 1.0
cdef C_FLOAT32 HALF_BOX_LENGTH = 0.5
cdef C_FLOAT32 FLOAT32_ZERO    = 0.0
cdef C_FLOAT32 FLOAT32_ONE     = 1.0
cdef C_FLOAT32 FLOAT_TWO       = 2.0
cdef C_INT32   INT32_ZERO      = 0
cdef C_INT32   INT32_ONE       = 1



@cython.nonecheck(False)
@cython.boundscheck(False)
@cython.wraparound(False)
@cython.cdivision(True)
@cython.always_allow_keywords(False)
cdef void _single_pairs_histograms( C_INT32          atomIndex, 
                                    C_INT32          atomSymbolIndex,
                                    C_INT32          atomMoleculeIndex,
                                    C_INT32          startIndex, 
                                    C_INT32          endIndex, 
                                    C_FLOAT32[:]     distances,
                                    C_INT32[:]       moleculeIndex,
                                    C_INT32[:]       elementIndex,
                                    C_FLOAT32[:,:,:] hintra,
                                    C_FLOAT32[:,:,:] hinter,
                                    C_FLOAT32        minDistance,
                                    C_FLOAT32        maxDistance,
                                    C_FLOAT32        bin,
                                    C_INT32          ncores = 1) nogil:
    cdef C_FLOAT32 distance
    cdef C_INT32 i, binIndex
    cdef C_INT32 num_threads = ncores
    for i in prange(startIndex, endIndex, INT32_ONE, nogil=True, schedule="static", num_threads=num_threads):
        if i == atomIndex: continue
        # get distance         
        distance = distances[i]
        # check limits
        if distance<minDistance:
            continue
        if distance>=maxDistance:
            continue
        # get index
        binIndex = <C_INT32>((distance-minDistance)/bin)
        # increment histograms
        if moleculeIndex[i] == atomMoleculeIndex:
            hintra[atomSymbolIndex,elementIndex[i],binIndex] += FLOAT32_ONE
        else:
            hinter[atomSymbolIndex,elementIndex[i],binIndex] += FLOAT32_ONE
    
                             

@cython.nonecheck(False)
@cython.boundscheck(False)
@cython.wraparound(False)
@cython.cdivision(True)
@cython.always_allow_keywords(False)
def single_pairs_histograms( C_INT32                    atomIndex, 
                             ndarray[C_FLOAT32, ndim=1] distances not None,
                             ndarray[C_INT32, ndim=1]   moleculeIndex not None,
                             ndarray[C_INT32, ndim=1]   elementIndex not None,
                             ndarray[C_FLOAT32, ndim=3] hintra not None,
                             ndarray[C_FLOAT32, ndim=3] hinter not None,
                             C_FLOAT32                  minDistance,
                             C_FLOAT32                  maxDistance,
                             C_FLOAT32                  bin, 
                             bint                       allAtoms = True,
                             C_INT32                    ncores = 1):
    """
    Computes the pair distribution histograms of a single atom given a distances array.
    
    :Arguments:
       #. atomIndex (int32): The index of the atom.
       #. distances (float32 array): The distances array.
       #. moleculeIndex (int32 array): The molecule's index array, assigning a molecule index for every atom.
       #. elementIndex (int32 array): The element's index array, assigning an element index for every atom.
       #. hintra (float32 array): The (numberOfElements,numberOfElements,1) array for intra-molecular distances histograms.
       #. hinter (float32 array): The (numberOfElements,numberOfElements,1) array for inter-molecular distances histograms.
       #. minDistance (float32): The minimum distance to be counted in the histogram.
       #. maxDistance (float32): The maximum distance to be counted in the histogram.
       #. bin (float32): The histogram bin size.
       #. allAtoms (bool): Perform the calculation over all the atoms. If False calculation starts from the given atomIndex. DEFAULT: True
       #. ncores (int32) [default=1]: The number of cores to use. 
                                  
    :Returns:
       #. hintra (float32 array): The updated (numberOfElements,numberOfElements,1) array for intra-molecular distances histograms.
       #. hinter (float32 array): The updated (numberOfElements,numberOfElements,1) array for inter-molecular distances histograms.
    """
    # declare variables
    cdef C_INT32 i, startIndex, endIndex
    cdef C_INT32 binIndex
    cdef C_INT32 atomMoleculeIndex, atomSymbolIndex
    cdef C_FLOAT32 float32Var, distance
    cdef C_INT32 num_threads = ncores
    # cast arguments
    bin = <C_FLOAT32>bin
    minDistance = <C_FLOAT32>minDistance
    maxDistance = <C_FLOAT32>maxDistance
    # get atom molecule and symbol
    atomMoleculeIndex = moleculeIndex[atomIndex]
    atomSymbolIndex   = elementIndex[atomIndex]
    # start index
    if allAtoms:
        startIndex = <C_INT32>0
    else:
        startIndex = <C_INT32>atomIndex
    endIndex = <C_INT32>distances.shape[0]
    # compute histograms
    _single_pairs_histograms( atomIndex         = atomIndex, 
                              atomSymbolIndex   = atomSymbolIndex,
                              atomMoleculeIndex = atomMoleculeIndex,
                              startIndex        = startIndex, 
                              endIndex          = endIndex, 
                              distances         = distances,
                              moleculeIndex     = moleculeIndex,
                              elementIndex      = elementIndex,
                              hintra            = hintra,
                              hinter            = hinter,
                              minDistance       = minDistance,
                              maxDistance       = maxDistance,
                              bin               = bin,
                              ncores            = ncores)

            

@cython.nonecheck(False)
@cython.boundscheck(False)
@cython.wraparound(False)
@cython.cdivision(True)
@cython.always_allow_keywords(False)
def multiple_pairs_histograms_coords( ndarray[C_INT32, ndim=1]      indexes not None,
                                      np.ndarray[C_FLOAT32, ndim=2] boxCoords not None,
                                      np.ndarray[C_FLOAT32, ndim=2] basis not None,
                                      bint                          isPBC,
                                      ndarray[C_INT32, ndim=1]      moleculeIndex not None,
                                      ndarray[C_INT32, ndim=1]      elementIndex not None,
                                      C_INT32                       numberOfElements,
                                      C_FLOAT32                     minDistance,
                                      C_FLOAT32                     maxDistance,
                                      C_FLOAT32                     bin,
                                      C_INT32                       histSize,
                                      bint                          allAtoms = True,
                                      C_INT32                       ncores = 1 ):    
    """
    Computes the pair distribution histograms of multiple atoms given atomic coordinates.
    
    :Arguments:
       #. indexes (int32 (k,3) numpy.ndarray): The atomic coordinates indexes array.
       #. coords (float32 (n,3) numpy.ndarray): The atomic coordinates array.
       #. basis (float32 (3,3) numpy.ndarray): The (3x3) boundary conditions box vectors.
       #. isPBC (bool): Whether it is a periodic boundary conditions or infinite.
       #. moleculeIndex (int32 array): The molecule's index array, assigning a molecule index for every atom.
       #. elementIndex (int32 array): The element's index array, assigning an element index for every atom.
       #. numberOfElements (int32): The number of elements in the system.
       #. minDistance (float32): The minimum distance to be counted in the histogram.
       #. maxDistance (float32): The maximum distance to be counted in the histogram.
       #. bin (float32): The histogram bin size.
       #. histSize(int32): The histograms size.
       #. allAtoms (bool): Perform the calculation over all the atoms. If False calculation starts from the given atomIndex. DEFAULT: True
       #. ncores (int32) [default=1]: The number of cores to use. 
       
    :Returns:
       #. hintra (float32 array): The updated (numberOfElements,numberOfElements,1) array for intra-molecular distances histograms.
       #. hinter (float32 array): The updated (numberOfElements,numberOfElements,1) array for inter-molecular distances histograms.
    """
    # declare variables
    cdef C_INT32 i, ii
    # cast arguments
    bin         = <C_FLOAT32>bin
    minDistance = <C_FLOAT32>minDistance
    maxDistance = <C_FLOAT32>maxDistance
    histSize    = <C_INT32>histSize
    # create histograms
    cdef ndarray[C_FLOAT32,  mode="c", ndim=3] hintra = np.zeros((numberOfElements,numberOfElements,histSize), dtype=NUMPY_FLOAT32)
    cdef ndarray[C_FLOAT32,  mode="c", ndim=3] hinter = np.zeros((numberOfElements,numberOfElements,histSize), dtype=NUMPY_FLOAT32)
    
    # loop atoms
    for i in indexes:
        # compute distances
        distances = pairs_distances_to_indexcoords( atomIndex = i, 
                                                    coords    = boxCoords,
                                                    basis     = basis,
                                                    isPBC     = isPBC,
                                                    allAtoms  = allAtoms,
                                                    ncores    = ncores)
        # compute histogram           
        single_pairs_histograms( atomIndex     = i, 
                                 distances     = distances,
                                 moleculeIndex = moleculeIndex,
                                 elementIndex  = elementIndex,
                                 hintra        = hintra,
                                 hinter        = hinter,
                                 minDistance   = minDistance,
                                 maxDistance   = maxDistance,
                                 bin           = bin,
                                 allAtoms      = allAtoms,
                                 ncores        = ncores )
    return hintra, hinter
    

@cython.nonecheck(False)
@cython.boundscheck(False)
@cython.wraparound(False)
@cython.cdivision(True)
@cython.always_allow_keywords(False)
def multiple_pairs_histograms_dists( ndarray[C_INT32, ndim=1]      indexes not None,
                                     np.ndarray[C_FLOAT32, ndim=2] distances not None,
                                     ndarray[C_INT32, ndim=1]      moleculeIndex not None,
                                     ndarray[C_INT32, ndim=1]      elementIndex not None,
                                     C_INT32                       numberOfElements,
                                     C_FLOAT32                     minDistance,
                                     C_FLOAT32                     maxDistance,
                                     C_FLOAT32                     bin,
                                     C_INT32                       histSize,
                                     bint                          allAtoms=True,
                                     C_INT32                       ncores = 1):    
    """
    Computes the pair distribution histograms of multiple atoms given atomic distances.
    
    :Arguments:
       #. indexes (int32 (k,3) numpy.ndarray): The atomic coordinates indexes array.
       #. distances (float32 array): The distances array.
       #. moleculeIndex (int32 array): The molecule's index array, assigning a molecule index for every atom.
       #. elementIndex (int32 array): The element's index array, assigning an element index for every atom.
       #. numberOfElements (int32): The number of elements in the system.
       #. minDistance (float32): The minimum distance to be counted in the histogram.
       #. maxDistance (float32): The maximum distance to be counted in the histogram.
       #. bin (float32): The histogram bin size.
       #. histSize(int32): The histograms size.
       #. allAtoms (bool): Perform the calculation over all the atoms. If False calculation starts from the given atomIndex. DEFAULT: True
       #. ncores (int32) [default=1]: The number of cores to use. 
       
    :Returns:
       #. hintra (float32 array): The updated (numberOfElements,numberOfElements,1) array for intra-molecular distances histograms.
       #. hinter (float32 array): The updated (numberOfElements,numberOfElements,1) array for inter-molecular distances histograms.
    """
    # declare variables
    cdef C_INT32 i
    # cast arguments
    bin         = <C_FLOAT32>bin
    minDistance = <C_FLOAT32>minDistance
    maxDistance = <C_FLOAT32>maxDistance
    histSize    = <C_INT32>histSize
    # create histograms
    cdef ndarray[C_FLOAT32,  mode="c", ndim=3] hintra = np.zeros((numberOfElements,numberOfElements,histSize), dtype=NUMPY_FLOAT32)
    cdef ndarray[C_FLOAT32,  mode="c", ndim=3] hinter = np.zeros((numberOfElements,numberOfElements,histSize), dtype=NUMPY_FLOAT32)

    # loop
    for i from <C_INT32>0 <= i < <C_INT32>indexes.shape[0]:
        # compute histogram   
        single_pairs_histograms( atomIndex     = indexes[i], 
                                 distances     = distances[:,i],
                                 moleculeIndex = moleculeIndex,
                                 elementIndex  = elementIndex,
                                 hintra        = hintra,
                                 hinter        = hinter,
                                 minDistance   = minDistance,
                                 maxDistance   = maxDistance,
                                 bin           = bin,
                                 allAtoms      = allAtoms,
                                 ncores        = ncores )
    return hintra, hinter
    
    
@cython.nonecheck(False)
@cython.boundscheck(False)
@cython.wraparound(False)
@cython.cdivision(True)
@cython.always_allow_keywords(False)
def full_pairs_histograms_coords( np.ndarray[C_FLOAT32, ndim=2] boxCoords not None,
                                  np.ndarray[C_FLOAT32, ndim=2] basis not None,
                                  bint                          isPBC,
                                  ndarray[C_INT32, ndim=1]      moleculeIndex not None,
                                  ndarray[C_INT32, ndim=1]      elementIndex not None,
                                  C_INT32                       numberOfElements,
                                  C_FLOAT32                     minDistance,
                                  C_FLOAT32                     maxDistance,
                                  C_FLOAT32                     bin,
                                  C_INT32                       histSize,
                                  C_INT32                       ncores = 1):    
    """
    Computes the pair distribution histograms of multiple atoms given atomic coordinates.
    
    :Arguments:
       #. boxCoords (float32 (n,3) numpy.ndarray): The atomic coordinates array.
       #. basis (float32 (3,3) numpy.ndarray): The (3x3) boundary conditions box vectors.
       #. isPBC (bool): Whether it is a periodic boundary conditions or infinite.
       #. moleculeIndex (int32 array): The molecule's index array, assigning a molecule index for every atom.
       #. elementIndex (int32 array): The element's index array, assigning an element index for every atom.
       #. numberOfElements (int32): The number of elements in the system.
       #. minDistance (float32): The minimum distance to be counted in the histogram.
       #. maxDistance (float32): The maximum distance to be counted in the histogram.
       #. bin (float32): The histogram bin size.
       #. histSize(int32): The histograms size.
       #. ncores (int32) [default=1]: The number of cores to use. 
       
    :Returns:
       #. hintra (float32 array): The updated (numberOfElements,numberOfElements,1) array for intra-molecular distances histograms.
       #. hinter (float32 array): The updated (numberOfElements,numberOfElements,1) array for inter-molecular distances histograms.
    """
    # get number of atoms
    cdef ndarray[C_INT32,  mode="c", ndim=1] indexes = np.arange(<C_INT32>boxCoords.shape[0], dtype=NUMPY_INT32)
    # calculate histograms
    return multiple_pairs_histograms_coords(indexes          = indexes,
                                            boxCoords        = boxCoords,
                                            basis            = basis,
                                            isPBC            = isPBC,
                                            moleculeIndex    = moleculeIndex,
                                            elementIndex     = elementIndex,
                                            numberOfElements = numberOfElements,
                                            minDistance      = minDistance,
                                            maxDistance      = maxDistance,
                                            bin              = bin,
                                            histSize         = histSize,
                                            ncores           = ncores,
                                            allAtoms         = False)


@cython.nonecheck(False)
@cython.boundscheck(False)
@cython.wraparound(False)
@cython.cdivision(True)
@cython.always_allow_keywords(False)
def full_pairs_histograms_dists( np.ndarray[C_FLOAT32, ndim=2] distances not None,
                                 ndarray[C_INT32, ndim=1]      moleculeIndex not None,
                                 ndarray[C_INT32, ndim=1]      elementIndex not None,
                                 C_INT32                       numberOfElements,
                                 C_FLOAT32                     minDistance,
                                 C_FLOAT32                     maxDistance,
                                 C_FLOAT32                     bin,
                                 C_INT32                       histSize,
                                 C_INT32                       ncores = 1):    
    """
    Computes the pair distribution histograms of multiple atoms given atomic distances.
    
    :Arguments:
       #. distances (float32 array): The distances array.
       #. moleculeIndex (int32 array): The molecule's index array, assigning a molecule index for every atom.
       #. elementIndex (int32 array): The element's index array, assigning an element index for every atom.
       #. numberOfElements (int32): The number of elements in the system.
       #. minDistance (float32): The minimum distance to be counted in the histogram.
       #. maxDistance (float32): The maximum distance to be counted in the histogram.
       #. bin (float32): The histogram bin size.
       #. histSize(int32): The histograms size.
       #. ncores (int32) [default=1]: The number of cores to use. 
       
    :Returns:
       #. hintra (float32 array): The updated (numberOfElements,numberOfElements,1) array for intra-molecular distances histograms.
       #. hinter (float32 array): The updated (numberOfElements,numberOfElements,1) array for inter-molecular distances histograms.
    """
    # get indexes
    cdef ndarray[C_INT32,  mode="c", ndim=1] indexes = np.arange(<C_INT32>distances.shape[1], dtype=NUMPY_INT32)
    # calculate histograms
    return multiple_pairs_histograms_dists(indexes          = indexes,
                                           distances        = distances,
                                           moleculeIndex    = moleculeIndex,
                                           elementIndex     = elementIndex,
                                           numberOfElements = numberOfElements,
                                           minDistance      = minDistance,
                                           maxDistance      = maxDistance,
                                           histSize         = histSize,
                                           bin              = bin,
                                           allAtoms         = False,
                                           ncores           = ncores)





                                        